# -*- coding: utf-8 -*-

# Copyright (C) 2025 Ross Scroggs All Rights Reserved.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""GAM command line argument processing

"""

class GamCLArgs():

# GAM entity types as specified on the command line
  ENTITY_BROWSER = 'ids'
  ENTITY_BROWSER_OU = 'browserou'
  ENTITY_BROWSER_OUS = 'browserous'
  ENTITY_BROWSER_QUERIES = 'browserqueies'
  ENTITY_BROWSER_QUERY = 'browserquery'
  ENTITY_CIGROUP = 'cigroup'
  ENTITY_CIGROUPS = 'cigroups'
  ENTITY_CIGROUP_USERS = 'cigroup_users'
  ENTITY_COURSEPARTICIPANTS = 'courseparticipants'
  ENTITY_CROS = 'cros'
  ENTITY_CROS_QUERIES = 'crosqueries'
  ENTITY_CROS_QUERY = 'crosquery'
  ENTITY_CROS_OU = 'cros_ou'
  ENTITY_CROS_OU_AND_CHILDREN = 'cros_ou_and_children'
  ENTITY_CROS_OUS = 'cros_ous'
  ENTITY_CROS_OUS_AND_CHILDREN = 'cros_ous_and_children'
  ENTITY_CROS_OU_QUERY = 'cros_ou_query'
  ENTITY_CROS_OU_AND_CHILDREN_QUERY = 'cros_ou_and_children_query'
  ENTITY_CROS_OUS_QUERY = 'cros_ous_query'
  ENTITY_CROS_OUS_AND_CHILDREN_QUERY = 'cros_ous_and_children_query'
  ENTITY_CROS_OU_QUERIES = 'cros_ou_queries'
  ENTITY_CROS_OU_AND_CHILDREN_QUERIES = 'cros_ou_and_children_queries'
  ENTITY_CROS_OUS_QUERIES = 'cros_ous_queries'
  ENTITY_CROS_OUS_AND_CHILDREN_QUERIES = 'cros_ous_and_children_queries'
  ENTITY_CROS_SN = 'cros_sn'
  ENTITY_DOMAINS = 'domains'
  ENTITY_DOMAINS_NS = 'domains_ns'
  ENTITY_DOMAINS_SUSP = 'domains_susp'
  ENTITY_GROUP = 'group'
  ENTITY_GROUP_INDE = 'group_inde'
  ENTITY_GROUP_NS = 'group_ns'
  ENTITY_GROUP_SUSP = 'group_susp'
  ENTITY_GROUPS = 'groups'
  ENTITY_GROUPS_INDE = 'groups_inde'
  ENTITY_GROUPS_NS = 'groups_ns'
  ENTITY_GROUPS_SUSP = 'groups_susp'
  ENTITY_GROUP_USERS = 'group_users'
  ENTITY_GROUP_USERS_NS = 'group_users_ns'
  ENTITY_GROUP_USERS_SUSP = 'group_users_susp'
  ENTITY_GROUP_USERS_SELECT = 'group_users_select'
  ENTITY_LICENSES = 'licenses'
  ENTITY_OAUTHUSER = 'oauthuser'
  ENTITY_OU = 'ou'
  ENTITY_OU_NS = 'ou_ns'
  ENTITY_OU_SUSP = 'ou_susp'
  ENTITY_OU_AND_CHILDREN = 'ou_and_children'
  ENTITY_OU_AND_CHILDREN_NS = 'ou_and_children_ns'
  ENTITY_OU_AND_CHILDREN_SUSP = 'ou_and_children_susp'
  ENTITY_OUS = 'ous'
  ENTITY_OUS_NS = 'ous_ns'
  ENTITY_OUS_SUSP = 'ous_susp'
  ENTITY_OUS_AND_CHILDREN = 'ous_and_children'
  ENTITY_OUS_AND_CHILDREN_NS = 'ous_and_children_ns'
  ENTITY_OUS_AND_CHILDREN_SUSP = 'ous_and_children_susp'
  ENTITY_QUERIES = 'queries'
  ENTITY_QUERY = 'query'
  ENTITY_STUDENTS = 'students'
  ENTITY_TEACHERS = 'teachers'
  ENTITY_USER = 'user'
  ENTITY_USERS = 'users'
  ENTITY_USERS_NS = 'users_ns'
  ENTITY_USERS_NS_SUSP = 'users_ns_susp'
  ENTITY_USERS_SUSP = 'users_susp'
#
  BROWSER_ENTITIES = [
    ENTITY_BROWSER,
    ENTITY_BROWSER_QUERIES,
    ENTITY_BROWSER_QUERY,
    ENTITY_BROWSER_OU,
    ENTITY_BROWSER_OUS,
    ]
  CROS_ENTITIES = [
    ENTITY_CROS,
    ENTITY_CROS_QUERIES,
    ENTITY_CROS_QUERY,
    ENTITY_CROS_OU,
    ENTITY_CROS_OU_AND_CHILDREN,
    ENTITY_CROS_OUS,
    ENTITY_CROS_OUS_AND_CHILDREN,
    ENTITY_CROS_OU_QUERY,
    ENTITY_CROS_OU_AND_CHILDREN_QUERY,
    ENTITY_CROS_OUS_QUERY,
    ENTITY_CROS_OUS_AND_CHILDREN_QUERY,
    ENTITY_CROS_OU_QUERIES,
    ENTITY_CROS_OU_AND_CHILDREN_QUERIES,
    ENTITY_CROS_OUS_QUERIES,
    ENTITY_CROS_OUS_AND_CHILDREN_QUERIES,
    ENTITY_CROS_SN,
    ]
  USER_ENTITIES = [
    ENTITY_CIGROUP,
    ENTITY_CIGROUPS,
    ENTITY_CIGROUP_USERS,
    ENTITY_COURSEPARTICIPANTS,
    ENTITY_DOMAINS,
    ENTITY_DOMAINS_NS,
    ENTITY_DOMAINS_SUSP,
    ENTITY_GROUP,
    ENTITY_GROUP_INDE,
    ENTITY_GROUP_NS,
    ENTITY_GROUP_SUSP,
    ENTITY_GROUPS,
    ENTITY_GROUPS_INDE,
    ENTITY_GROUPS_NS,
    ENTITY_GROUPS_SUSP,
    ENTITY_GROUP_USERS,
    ENTITY_GROUP_USERS_NS,
    ENTITY_GROUP_USERS_SUSP,
    ENTITY_GROUP_USERS_SELECT,
    ENTITY_LICENSES,
    ENTITY_OAUTHUSER,
    ENTITY_OU,
    ENTITY_OU_NS,
    ENTITY_OU_SUSP,
    ENTITY_OU_AND_CHILDREN,
    ENTITY_OU_AND_CHILDREN_NS,
    ENTITY_OU_AND_CHILDREN_SUSP,
    ENTITY_OUS,
    ENTITY_OUS_NS,
    ENTITY_OUS_SUSP,
    ENTITY_OUS_AND_CHILDREN,
    ENTITY_OUS_AND_CHILDREN_NS,
    ENTITY_OUS_AND_CHILDREN_SUSP,
    ENTITY_QUERIES,
    ENTITY_QUERY,
    ENTITY_STUDENTS,
    ENTITY_TEACHERS,
    ENTITY_USER,
    ENTITY_USERS,
    ]
# Aliases for CL entity types
  ENTITY_ALIAS_MAP = {
    'browsers': ENTITY_BROWSER,
    'browserorg': ENTITY_BROWSER_OU,
    'browserorgs': ENTITY_BROWSER_OUS,
    'crosorg': ENTITY_CROS_OU,
    'crosorg_and_child': ENTITY_CROS_OU_AND_CHILDREN,
    'crosorg_and_children': ENTITY_CROS_OU_AND_CHILDREN,
    'crosorgs': ENTITY_CROS_OUS,
    'crosorgs_and_child': ENTITY_CROS_OUS_AND_CHILDREN,
    'crosorgs_and_children': ENTITY_CROS_OUS_AND_CHILDREN,
    'crosou_and_child': ENTITY_CROS_OU_AND_CHILDREN,
    'crosou_and_childen': ENTITY_CROS_OU_AND_CHILDREN,
    'crosous_and_child': ENTITY_CROS_OUS_AND_CHILDREN,
    'crosous_and_children': ENTITY_CROS_OUS_AND_CHILDREN,
    'cros_org': ENTITY_CROS_OU,
    'cros_org_and_child': ENTITY_CROS_OU_AND_CHILDREN,
    'cros_org_and_children': ENTITY_CROS_OU_AND_CHILDREN,
    'cros_orgs': ENTITY_CROS_OUS,
    'cros_orgs_and_child': ENTITY_CROS_OUS_AND_CHILDREN,
    'cros_orgs_and_children': ENTITY_CROS_OUS_AND_CHILDREN,
    'cros_ou_and_child': ENTITY_CROS_OU_AND_CHILDREN,
    'cros_ou_and_childen': ENTITY_CROS_OU_AND_CHILDREN,
    'cros_ous_and_child': ENTITY_CROS_OUS_AND_CHILDREN,
    'cros_ous_and_children': ENTITY_CROS_OUS_AND_CHILDREN,
    'crosorg_query': ENTITY_CROS_OU_QUERY,
    'crosorg_and_child_query': ENTITY_CROS_OU_AND_CHILDREN_QUERY,
    'crosorg_and_children_query': ENTITY_CROS_OU_AND_CHILDREN_QUERY,
    'crosorgs_query': ENTITY_CROS_OUS_QUERY,
    'crosorgs_and_child_query': ENTITY_CROS_OUS_AND_CHILDREN_QUERY,
    'crosorgs_and_children_query': ENTITY_CROS_OUS_AND_CHILDREN_QUERY,
    'crosou_and_child_query': ENTITY_CROS_OU_AND_CHILDREN_QUERY,
    'crosou_and_childen_query': ENTITY_CROS_OU_AND_CHILDREN_QUERY,
    'crosous_and_child_query': ENTITY_CROS_OUS_AND_CHILDREN_QUERY,
    'crosous_and_children_query': ENTITY_CROS_OUS_AND_CHILDREN_QUERY,
    'cros_org_query': ENTITY_CROS_OU_QUERY,
    'cros_org_and_child_query': ENTITY_CROS_OU_AND_CHILDREN_QUERY,
    'cros_org_and_children_query': ENTITY_CROS_OU_AND_CHILDREN_QUERY,
    'cros_orgs_query': ENTITY_CROS_OUS_QUERY,
    'cros_orgs_and_child_query': ENTITY_CROS_OUS_AND_CHILDREN_QUERY,
    'cros_orgs_and_children_query': ENTITY_CROS_OUS_AND_CHILDREN_QUERY,
    'cros_ou_and_child_query': ENTITY_CROS_OU_AND_CHILDREN_QUERY,
    'cros_ou_and_childen_query': ENTITY_CROS_OU_AND_CHILDREN_QUERY,
    'cros_ous_and_child_query': ENTITY_CROS_OUS_AND_CHILDREN_QUERY,
    'cros_ous_and_children_query': ENTITY_CROS_OUS_AND_CHILDREN_QUERY,
    'crosorg_queries': ENTITY_CROS_OU_QUERIES,
    'crosorg_and_child_queries': ENTITY_CROS_OU_AND_CHILDREN_QUERIES,
    'crosorg_and_children_queries': ENTITY_CROS_OU_AND_CHILDREN_QUERIES,
    'crosorgs_queries': ENTITY_CROS_OUS_QUERIES,
    'crosorgs_and_child_queries': ENTITY_CROS_OUS_AND_CHILDREN_QUERIES,
    'crosorgs_and_children_queries': ENTITY_CROS_OUS_AND_CHILDREN_QUERIES,
    'crosou_and_child_queries': ENTITY_CROS_OU_AND_CHILDREN_QUERIES,
    'crosou_and_childen_queries': ENTITY_CROS_OU_AND_CHILDREN_QUERIES,
    'crosous_and_child_queries': ENTITY_CROS_OUS_AND_CHILDREN_QUERIES,
    'crosous_and_children_queries': ENTITY_CROS_OUS_AND_CHILDREN_QUERIES,
    'cros_org_queries': ENTITY_CROS_OU_QUERIES,
    'cros_org_and_child_queries': ENTITY_CROS_OU_AND_CHILDREN_QUERIES,
    'cros_org_and_children_queries': ENTITY_CROS_OU_AND_CHILDREN_QUERIES,
    'cros_orgs_queries': ENTITY_CROS_OUS_QUERIES,
    'cros_orgs_and_child_queries': ENTITY_CROS_OUS_AND_CHILDREN_QUERIES,
    'cros_orgs_and_children_queries': ENTITY_CROS_OUS_AND_CHILDREN_QUERIES,
    'cros_ou_and_child_queries': ENTITY_CROS_OU_AND_CHILDREN_QUERIES,
    'cros_ou_and_childen_queries': ENTITY_CROS_OU_AND_CHILDREN_QUERIES,
    'cros_ous_and_child_queries': ENTITY_CROS_OUS_AND_CHILDREN_QUERIES,
    'cros_ous_and_children_queries': ENTITY_CROS_OUS_AND_CHILDREN_QUERIES,
    'license': ENTITY_LICENSES,
    'licence': ENTITY_LICENSES,
    'licences': ENTITY_LICENSES,
    'org': ENTITY_OU,
    'org_ns': ENTITY_OU_NS,
    'org_susp': ENTITY_OU_SUSP,
    'org_and_child': ENTITY_OU_AND_CHILDREN,
    'org_and_child_ns': ENTITY_OU_AND_CHILDREN_NS,
    'org_and_child_susp': ENTITY_OU_AND_CHILDREN_SUSP,
    'org_and_children': ENTITY_OU_AND_CHILDREN,
    'org_and_children_ns': ENTITY_OU_AND_CHILDREN_NS,
    'org_and_children_susp': ENTITY_OU_AND_CHILDREN_SUSP,
    'orgs': ENTITY_OUS,
    'orgs_ns': ENTITY_OUS_NS,
    'orgs_susp': ENTITY_OUS_SUSP,
    'orgs_and_child': ENTITY_OUS_AND_CHILDREN,
    'orgs_and_child_ns': ENTITY_OUS_AND_CHILDREN_NS,
    'orgs_and_child_susp': ENTITY_OUS_AND_CHILDREN_SUSP,
    'orgs_and_children': ENTITY_OUS_AND_CHILDREN,
    'orgs_and_children_ns': ENTITY_OUS_AND_CHILDREN_NS,
    'orgs_and_children_susp': ENTITY_OUS_AND_CHILDREN_SUSP,
    'ou_and_child': ENTITY_OU_AND_CHILDREN,
    'ou_and_child_ns': ENTITY_OU_AND_CHILDREN_NS,
    'ou_and_child_susp': ENTITY_OU_AND_CHILDREN_SUSP,
    'ous_and_child': ENTITY_OUS_AND_CHILDREN,
    'ous_and_child_ns': ENTITY_OUS_AND_CHILDREN_NS,
    'ous_and_child_susp': ENTITY_OUS_AND_CHILDREN_SUSP,
    }
# CL entity source selectors
  ENTITY_SELECTOR_ALL = 'all'
  ENTITY_SELECTOR_CSV = 'csv'
  ENTITY_SELECTOR_CSVDATAFILE = 'csvdatafile'
  ENTITY_SELECTOR_CSVFILE = 'csvfile'
  ENTITY_SELECTOR_FILE = 'file'
  ENTITY_SELECTOR_DATAFILE = 'datafile'
  ENTITY_SELECTOR_CROSCSV = 'croscsv'
  ENTITY_SELECTOR_CROSCSV_SN = 'croscsv_sn'
  ENTITY_SELECTOR_CROSCSVFILE = 'croscsvfile'
  ENTITY_SELECTOR_CROSCSVFILE_SN = 'croscsvfile_sn'
  ENTITY_SELECTOR_CROSFILE = 'crosfile'
  ENTITY_SELECTOR_CROSFILE_SN = 'crosfile_sn'
  ENTITY_SELECTOR_CSVKMD = 'csvkmd'
  ENTITY_SELECTOR_CSVSUBKEY = 'csvsubkey'
  ENTITY_SELECTOR_CSVDATA = 'csvdata'
  ENTITY_SELECTOR_CROSCSVDATA = 'croscsvdata'
#
  SERVICE_ACCOUNT_ONLY_ENTITY_SELECTORS = [
    ENTITY_SELECTOR_CSVDATAFILE,
    ENTITY_SELECTOR_CSVKMD,
    ENTITY_SELECTOR_CSVSUBKEY,
    ENTITY_SELECTOR_DATAFILE,
    ]
  ENTITY_LIST_SELECTORS = [
    ENTITY_SELECTOR_CSVKMD,
    ENTITY_SELECTOR_CSVSUBKEY,
    ENTITY_SELECTOR_CSV,
    ENTITY_SELECTOR_CSVFILE,
    ENTITY_SELECTOR_FILE,
    ENTITY_SELECTOR_CSVDATA,
    ]
  BASE_ENTITY_SELECTORS = [
    ENTITY_SELECTOR_ALL,
    ENTITY_SELECTOR_CSVDATAFILE,
    ENTITY_SELECTOR_CSVKMD,
    ENTITY_SELECTOR_CSVSUBKEY,
    ENTITY_SELECTOR_DATAFILE,
    ]
  BROWSER_ENTITY_SELECTORS = [
    ENTITY_SELECTOR_CSV,
    ENTITY_SELECTOR_CSVFILE,
    ENTITY_SELECTOR_FILE,
    ]
  CROS_ENTITY_SELECTORS = [
    ENTITY_SELECTOR_CROSCSV,
    ENTITY_SELECTOR_CROSCSV_SN,
    ENTITY_SELECTOR_CROSCSVFILE,
    ENTITY_SELECTOR_CROSCSVFILE_SN,
    ENTITY_SELECTOR_CROSFILE,
    ENTITY_SELECTOR_CROSFILE_SN,
    ]
  USER_ENTITY_SELECTORS = [
    ENTITY_SELECTOR_CSV,
    ENTITY_SELECTOR_CSVFILE,
    ENTITY_SELECTOR_FILE,
    ]
  CROS_CSVDATA_ENTITY_SELECTORS = [
    ENTITY_SELECTOR_CROSCSVDATA,
    ]
  USER_CSVDATA_ENTITY_SELECTORS = [
    ENTITY_SELECTOR_CSVDATA,
    ]
# Allowed values for CL source selector all
  CROS_ENTITY_SELECTOR_ALL_SUBTYPES = [
    ENTITY_CROS,
    ]
  USER_ENTITY_SELECTOR_ALL_SUBTYPES = [
    ENTITY_USERS,
    ENTITY_USERS_NS,
    ENTITY_USERS_NS_SUSP,
    ENTITY_USERS_SUSP,
    ]
#
  ENTITY_ALL_CROS = ENTITY_SELECTOR_ALL+' '+ENTITY_CROS
  ENTITY_ALL_USERS = ENTITY_SELECTOR_ALL+' '+ENTITY_USERS
  ENTITY_ALL_USERS_NS = ENTITY_SELECTOR_ALL+' '+ENTITY_USERS_NS
  ENTITY_ALL_USERS_NS_SUSP = ENTITY_SELECTOR_ALL+' '+ENTITY_USERS_NS_SUSP
  ENTITY_ALL_USERS_SUSP = ENTITY_SELECTOR_ALL+' '+ENTITY_USERS_SUSP
#
  ALL_USERS_QUERY_MAP = {
    ENTITY_ALL_USERS: 'isSuspended=False',
    ENTITY_ALL_USERS_NS: 'isSuspended=False',
    ENTITY_ALL_USERS_NS_SUSP: None,
    ENTITY_ALL_USERS_SUSP: 'isSuspended=True',
  }
#
  ENTITY_SELECTOR_ALL_SUBTYPES_MAP = {
    ENTITY_CROS: ENTITY_ALL_CROS,
    ENTITY_USERS: ENTITY_ALL_USERS,
    ENTITY_USERS_NS: ENTITY_ALL_USERS_NS,
    ENTITY_USERS_NS_SUSP: ENTITY_ALL_USERS_NS_SUSP,
    ENTITY_USERS_SUSP: ENTITY_ALL_USERS_SUSP,
    }
# Allowed values for CL source selector datafile, csvkmd
  CROS_ENTITY_SELECTOR_DATAFILE_CSVKMD_SUBTYPES = [
    ENTITY_CROS,
    ENTITY_CROS_OUS,
    ENTITY_CROS_OUS_AND_CHILDREN,
    ENTITY_CROS_SN,
    ]
  USER_ENTITY_SELECTOR_DATAFILE_CSVKMD_SUBTYPES = [
    ENTITY_USERS,
    ENTITY_CIGROUPS,
    ENTITY_CIGROUP_USERS,
    ENTITY_DOMAINS,
    ENTITY_DOMAINS_NS,
    ENTITY_DOMAINS_SUSP,
    ENTITY_GROUPS,
    ENTITY_GROUPS_INDE,
    ENTITY_GROUPS_NS,
    ENTITY_GROUPS_SUSP,
    ENTITY_GROUP_USERS,
    ENTITY_GROUP_USERS_NS,
    ENTITY_GROUP_USERS_SUSP,
    ENTITY_GROUP_USERS_SELECT,
    ENTITY_OUS,
    ENTITY_OUS_NS,
    ENTITY_OUS_SUSP,
    ENTITY_OUS_AND_CHILDREN,
    ENTITY_OUS_AND_CHILDREN_NS,
    ENTITY_OUS_AND_CHILDREN_SUSP,
    ENTITY_COURSEPARTICIPANTS,
    ENTITY_STUDENTS,
    ENTITY_TEACHERS,
    ]
# Batch file commands
  EXECUTE_CMD = 'execute'
  GAM_CMD = 'gam'
  COMMIT_BATCH_CMD = 'commit-batch'
  PRINT_CMD = 'print'
  SET_CMD = 'set'
  CLEAR_CMD = 'clear'
  SLEEP_CMD = 'sleep'
# Command line batch/csv/loop/tbatch keywords
  BATCH_CMD = 'batch'
  CSV_CMD = 'csv'
  CSVTEST_CMD = 'csvtest'
  LOOP_CMD = 'loop'
  TBATCH_CMD = 'tbatch'
# Command line select/showsections/selectfilter/selectoutputfilter/selectinputfilter/config/redirect arguments
  SELECT_CMD = 'select'
  SHOWSECTIONS_CMD = 'showsections'
  SELECTFILTER_CMD = 'selectfilter'
  SELECTOUTPUTFILTER_CMD = 'selectoutputfilter'
  SELECTINPUTFILTER_CMD = 'selectinputfilter'
  CONFIG_CMD = 'config'
  MULTIPROCESSEXIT_CMD = 'multiprocessexit'
  REDIRECT_CMD = 'redirect'
  GAM_META_COMMANDS = [SELECT_CMD, SHOWSECTIONS_CMD, SELECTFILTER_CMD, SELECTOUTPUTFILTER_CMD, SELECTINPUTFILTER_CMD,
                       CONFIG_CMD, MULTIPROCESSEXIT_CMD, REDIRECT_CMD]
# Command line arguments
  ARG_3LO = '3lo'
  ARG_ACL = 'acl'
  ARG_ACLS = 'acls'
  ARG_ADDRESSES = 'addresses'
  ARG_ADMIN = 'admin'
  ARG_ADMINS = 'admins'
  ARG_ADMINROLE = 'adminrole'
  ARG_ADMINROLES = 'adminroles'
  ARG_ALERT = 'alert'
  ARG_ALERTS = 'alerts'
  ARG_ALERTFEEDBACK = 'alertfeedback'
  ARG_ALERTFEEDBACKS = 'alertfeedbacks'
  ARG_ALERTSFEEDBACK = 'alertsfeedback'
  ARG_ALIAS = 'alias'
  ARG_ALIASES = 'aliases'
  ARG_ALIASDOMAIN = 'aliasdomain'
  ARG_ALIASDOMAINS = 'aliasdomains'
  ARG_ANALYTICACCOUNT = 'analyticaccount'
  ARG_ANALYTICACCOUNTS = 'analyticaccounts'
  ARG_ANALYTICACCOUNTSUMMARY = 'analyticaccountsummary'
  ARG_ANALYTICACCOUNTSUMMARIES = 'analyticaccountsummaries'
  ARG_ANALYTICDATASTREAM = 'analyticdatastream'
  ARG_ANALYTICDATASTREAMS = 'analyticdatastreams'
  ARG_ANALYTICPROPERTY = 'analyticproperty'
  ARG_ANALYTICPROPERTIES = 'analyticproperties'
  ARG_API = 'api'
  ARG_APIS = 'apis'
  ARG_APIPROJECT = 'apiproject'
  ARG_APPACCESSSETTINGS = 'appaccesssettings'
  ARG_APPDETAILS = 'appdetails'
  ARG_APPLICATIONSPECIFICPASSWORDS = 'applicationspecificpasswords'
  ARG_ASP = 'asp'
  ARG_ASPS = 'asps'
  ARG_BACKUPCODE = 'backupcode'
  ARG_BACKUPCODES = 'backupcodes'
  ARG_BROWSER = 'browser'
  ARG_BROWSERS = 'browsers'
  ARG_BROWSERTOKEN = 'browsertoken'
  ARG_BROWSERTOKENS = 'browsertokens'
  ARG_BUCKET = 'bucket'
  ARG_BUCKETS = 'buckets'
  ARG_BUILDING = 'building'
  ARG_BUILDINGS = 'buildings'
  ARG_CAALEVEL = 'caalevel'
  ARG_CAALEVELS = 'caalevels'
  ARG_CALATTENDEES = 'calattendees'
  ARG_CALENDAR = 'calendar'
  ARG_CALENDARS = 'calendars'
  ARG_CALENDARACL = 'calendaracl'
  ARG_CALENDARACLS = 'calendaracls'
  ARG_CALENDARTRASH = 'calendartrash'
  ARG_CALSETTINGS = 'calsettings'
  ARG_CHANNELCUSTOMER = 'channelcustomer'
  ARG_CHANNELCUSTOMERS = 'channelcustomers'
  ARG_CHANNELCUSTOMERENTITLEMENT = 'channelcustomerentitlement'
  ARG_CHANNELCUSTOMERENTITLEMENTS = 'channelcustomerentitlements'
  ARG_CHANNELOFFER = 'channeloffer'
  ARG_CHANNELOFFERS = 'channeloffers'
  ARG_CHANNELPRODUCT = 'channelproduct'
  ARG_CHANNELPRODUCTS = 'channelproducts'
  ARG_CHANNELSKU = 'channelsku'
  ARG_CHANNELSKUS = 'channelskus'
  ARG_CHAT = 'chat'
  ARG_CHATEVENT = 'chatevent'
  ARG_CHATEVENTS = 'chatevents'
  ARG_CHATMEMBER = 'chatmember'
  ARG_CHATMEMBERS = 'chatmembers'
  ARG_CHATMESSAGE = 'chatmessage'
  ARG_CHATMESSAGES = 'chatmessages'
  ARG_CHATSPACE = 'chatspace'
  ARG_CHATSPACES = 'chatspaces'
  ARG_CHATSPACEDM = 'chatspacedm'
  ARG_CHROMEAPP = 'chromeapp'
  ARG_CHROMEAPPS = 'chromeapps'
  ARG_CHROMEAPPDEVICES = 'chromeappdevices'
  ARG_CHROMEAUES = 'chromeaues'
  ARG_CHROMEHISTORY = 'chromehistory'
  ARG_CHROMENEEDSATTN = 'chromeneedsattn'
  ARG_CHROMENETWORK = 'chromenetwork'
  ARG_CHROMENETWORKS = 'chromenetworks'
  ARG_CHROMEPOLICYIMAGE = 'chromepolicyimage'
  ARG_CHROMEPOLICY = 'chromepolicy'
  ARG_CHROMEPOLICIES = 'chromepolicies'
  ARG_CHROMEPROFILE = 'chromeprofile'
  ARG_CHROMEPROFILES = 'chromeprofiles'
  ARG_CHROMESCHEMA = 'chromeschema'
  ARG_CHROMESCHEMAS = 'chromeschemas'
  ARG_CHROMESNVALIDITY = 'chromesnvalidity'
  ARG_CHROMEVERSIONS = 'chromeversions'
  ARG_CIGROUP = 'cigroup'
  ARG_CIGROUPS = 'cigroups'
  ARG_CIGROUPMEMBERS = 'cigroupmembers'
  ARG_CIGROUPSMEMBERS = 'cigroupsmembers'
  ARG_CIMEMBER = 'cimember'
  ARG_CIMEMBERS = 'cimembers'
  ARG_CIPOLICY = 'policy'
  ARG_CIPOLICIES = 'policies'
  ARG_CLASSIFICATIONLABEL = 'classificationlabel'
  ARG_CLASSIFICATIONLABELS = 'classificationlabels'
  ARG_CLASSIFICATIONLABELPERMISSION = 'classificationlabelpermission'
  ARG_CLASSIFICATIONLABELPERMISSIONS = 'classificationlabelpermissions'
  ARG_CLASS = 'class'
  ARG_CLASSES = 'classes'
  ARG_CLASSPARTICIPANTS = 'classparticipants'
  ARG_CLASSROOMINVITATION = 'classroominvitation'
  ARG_CLASSROOMINVITATIONS = 'classroominvitations'
  ARG_CLASSROOMOAUTH2 = 'classroomoauth2'
  ARG_CLASSROOMPROFILE = 'classroomprofile'
  ARG_CONTACT = 'contact'
  ARG_CONTACTS = 'contacts'
  ARG_CONTACTDELEGATE = 'contactdelegate'
  ARG_CONTACTDELEGATES = 'contactdelegates'
  ARG_CONTACTGROUP = 'contactgroup'
  ARG_CONTACTGROUPS = 'contactgroups'
  ARG_CONTACTPHOTO = 'contactphoto'
  ARG_CONTACTPHOTOS = 'contactphotos'
  ARG_COUNT = 'count'
  ARG_COUNTS = 'counts'
  ARG_COURSE = 'course'
  ARG_COURSES = 'courses'
  ARG_COURSEANNOUNCEMENTS = 'courseannouncements'
  ARG_COURSEMATERIALS = 'coursematerials'
  ARG_COURSEPARTICIPANTS = 'courseparticipants'
  ARG_COURSESUBMISSIONS = 'coursesubmissions'
  ARG_COURSETOPICS = 'coursetopics'
  ARG_COURSEWORK = 'coursework'
  ARG_CROS = 'cros'
  ARG_CROSES = 'croses'
  ARG_CROSACTIVITY = 'crosactivity'
  ARG_CROSTELEMETRY = 'crostelemetry'
  ARG_CSEIDENTITY = 'cseidentity'
  ARG_CSEIDENTITIES = 'cseidentities'
  ARG_CSEKEYPAIR = 'csekeypair'
  ARG_CSEKEYPAIRS = 'csekeypairs'
  ARG_CURRENTPROJECTID = 'currentprojectid'
  ARG_CUSTOMER = 'customer'
  ARG_DATASTUDIOASSET = 'datastudioasset'
  ARG_DATASTUDIOASSETS = 'datastudioassets'
  ARG_DATASTUDIOPERMISSION = 'datastudiopermission'
  ARG_DATASTUDIOPERMISSIONS = 'datastudiopermissions'
  ARG_LOOKERSTUDIOASSET = 'lookerstudioasset'
  ARG_LOOKERSTUDIOASSETS = 'lookerstudioassets'
  ARG_LOOKERSTUDIOPERMISSION = 'lookerstudiopermission'
  ARG_LOOKERSTUDIOPERMISSIONS = 'lookerstudiopermissions'
  ARG_DATATRANSFER = 'datatransfer'
  ARG_DATATRANSFERS = 'datatransfers'
  ARG_DELEGATE = 'delegate'
  ARG_DELEGATES = 'delegates'
  ARG_DEVICE = 'device'
  ARG_DEVICES = 'devices'
  ARG_DEVICEFILE = 'devicefile'
  ARG_DEVICEFILES = 'devicefiles'
  ARG_DEVICEUSER = 'deviceuser'
  ARG_DEVICEUSERS = 'deviceusers'
  ARG_DEVICEUSERSTATE = 'deviceuserstate'
  ARG_DISKUSAGE = 'diskusage'
  ARG_DOCUMENT = 'document'
  ARG_DOMAIN = 'domain'
  ARG_DOMAINS = 'domains'
  ARG_DOMAINALIAS = 'domainalias'
  ARG_DOMAINALIASES = 'domainaliases'
  ARG_DOMAINCONTACT = 'domaincontact'
  ARG_DOMAINCONTACTS = 'domaincontacts'
  ARG_DOMAINCONTACTPHOTO = 'domaincontactphoto'
  ARG_DOMAINCONTACTPHOTOS = 'domaincontactphotas'
  ARG_DOMAINPROFILE = 'domainprofile'
  ARG_DOMAINPROFILES = 'domainprofiles'
  ARG_DRIVE = 'drive'
  ARG_DRIVEACTIVITY = 'driveactivity'
  ARG_DRIVEFILE = 'drivefile'
  ARG_DRIVEFILEACL = 'drivefileacl'
  ARG_DRIVEFILEACLS = 'drivefileacls'
  ARG_DRIVEFILESHORTCUT = 'drivefileshortcut'
  ARG_DRIVEFILESHORTCUTS = 'drivefileshortcuts'
  ARG_DRIVEFOLDERPATH = 'drivefolderpath'
  ARG_DRIVELABEL = 'drivelabel'
  ARG_DRIVELABELS = 'drivelabels'
  ARG_DRIVELABELPERMISSION = 'drivelabelpermission'
  ARG_DRIVELABELPERMISSIONS = 'drivelabelpermissions'
  ARG_DRIVELASTMODIFICATION = 'drivelastmodification'
  ARG_DRIVELASTMODIFICATIONS = 'drivelastmodifications'
  ARG_DRIVESETTINGS = 'drivesettings'
  ARG_DRIVETRASH = 'drivetrash'
  ARG_EMPTYDRIVEFOLDERS = 'emptydrivefolders'
  ARG_EVENT = 'event'
  ARG_EVENTS = 'events'
  ARG_EXPORT = 'export'
  ARG_EXPORTS = 'exports'
  ARG_FEATURE = 'feature'
  ARG_FEATURES = 'features'
  ARG_FILECOMMENT = 'filecomment'
  ARG_FILECOMMENTS = 'filecomments'
  ARG_FILECOUNT = 'filecount'
  ARG_FILECOUNTS = 'filecounts'
  ARG_FILEDRIVELABEL = 'filedrivelabel'
  ARG_FILEDRIVELABELS = 'filedrivelabels'
  ARG_FILEINFO = 'fileinfo'
  ARG_FILELIST = 'filelist'
  ARG_FILEPARENTTREE = 'fileparenttree'
  ARG_FILEPATH = 'filepath'
  ARG_FILEPATHS = 'filepaths'
  ARG_FILEREVISION = 'filerevision'
  ARG_FILEREVISIONS = 'filerevisions'
  ARG_FILESHARECOUNT = 'filesharecount'
  ARG_FILESHARECOUNTS = 'filesharecounts'
  ARG_FILETREE = 'filetree'
  ARG_FILTER = 'filter'
  ARG_FILTERS = 'filters'
  ARG_FOCUSTIME = 'focustime'
  ARG_FOCUSTIMES = 'focustimes'
  ARG_FORM = 'form'
  ARG_FORMS = 'forms'
  ARG_FORMRESPONSE = 'formresponse'
  ARG_FORMRESPONSES = 'formresponses'
  ARG_FORWARD = 'forward'
  ARG_FORWARDS = 'forwards'
  ARG_FORWARDINGADDRESS = 'forwardingaddress'
  ARG_FORWARDINGADDRESSES = 'forwardingaddresses'
  ARG_GCPFOLDER = 'gcpfolder'
  ARG_GCPSERVICEACCOUNT = 'gcpserviceaccount'
  ARG_GMAIL = 'gmail'
  ARG_GMAILPROFILE = 'gmailprofile'
  ARG_GROUP = 'group'
  ARG_GROUPS = 'groups'
  ARG_GROUPLIST = 'grouplist'
  ARG_GROUPSLIST = 'groupslist'
  ARG_GROUPMEMBERS = 'groupmembers'
  ARG_GROUPSMEMBERS = 'groupsmembers'
  ARG_GROUPTREE = 'grouptree'
  ARG_GUARDIAN = 'guardian'
  ARG_GUARDIANS = 'guardians'
  ARG_GUARDIANINVITE = 'guardianinvite'
  ARG_GUARDIANINVITATION = 'guardianinvitation'
  ARG_GUARDIANINVITATIONS = 'guardianinvitations'
  ARG_HOLD = 'hold'
  ARG_HOLDS = 'holds'
  ARG_IMAP = 'imap'
  ARG_IMAP4 = 'imap4'
  ARG_INBOUNDSSOASSIGNMENT = 'inboundssoassignment'
  ARG_INBOUNDSSOASSIGNMENTS = 'inboundssoassignments'
  ARG_INBOUNDSSOCREDENTIAL = 'inboundssocredential'
  ARG_INBOUNDSSOCREDENTIALS = 'inboundssocredentials'
  ARG_INBOUNDSSOPROFILE = 'inboundssoprofile'
  ARG_INBOUNDSSOPROFILES = 'inboundssoprofiles'
  ARG_INSTANCE = 'instance'
  ARG_INVITEGUARDIAN = 'inviteguardian'
  ARG_ISINVITABLE = 'isinvitable'
  ARG_LABEL = 'label'
  ARG_LABELS = 'labels'
  ARG_LABELLIST = 'labellist'
  ARG_LABELID = 'labelid'
  ARG_LABELIDLIST = 'labelidlist'
  ARG_LABELSETTINGS = 'labelsettings'
  ARG_LANGUAGE = 'language'
  ARG_LICENCE = 'licence'
  ARG_LICENCES = 'licences'
  ARG_LICENSE = 'license'
  ARG_LICENSES = 'licenses'
  ARG_MATTER = 'matter'
  ARG_MATTERS = 'matters'
  ARG_MEETSPACE = 'meetspace'
  ARG_MEETSPACES = 'meetspaces'
  ARG_MEETCONFERENCE = 'meetconference'
  ARG_MEETCONFERENCES = 'meetconferences'
  ARG_MEETPARTICIPANT = 'meetparticipant'
  ARG_MEETPARTICIPANTS = 'meetparticipants'
  ARG_MEETRECORDING = 'meetrecording'
  ARG_MEETRECORDINGS = 'meetrecordings'
  ARG_MEETTRANSCRIPT = 'meettranscript'
  ARG_MEETTRANSCRIPTS = 'meettranscripts'
  ARG_MEMBER = 'member'
  ARG_MEMBERS = 'members'
  ARG_MESSAGE = 'message'
  ARG_MESSAGES = 'messages'
  ARG_MESSAGEHISTORY = 'messagehistory'
  ARG_MOBILE = 'mobile'
  ARG_MOBILES = 'mobiles'
  ARG_NICKNAME = 'nickname'
  ARG_NICKNAMES = 'nicknames'
  ARG_NOTE = 'note'
  ARG_NOTES = 'notes'
  ARG_NOTEACL = 'noteacl'
  ARG_NOTEACLS = 'noteacls'
  ARG_NOTESACL = 'notesacl'
  ARG_NOTESACLS = 'notesacls'
  ARG_NOTEATTACHMENT = 'noteattachment'
  ARG_NOTEATTACHMENTS = 'noteattachments'
  ARG_OAUTH = 'oauth'
  ARG_ORG = 'org'
  ARG_ORGS = 'orgs'
  ARG_ORGTREE = 'orgtree'
  ARG_ORGUNIT = 'orgunit'
  ARG_ORGUNITS = 'orgunits'
  ARG_ORGUNITSHAREDDRIVE = 'orgunitshareddrive'
  ARG_ORGUNITSHAREDDRIVES = 'orgunitshareddrives'
  ARG_ORPHANS = 'orphans'
  ARG_OTHERCONTACT = 'othercontact'
  ARG_OTHERCONTACTS = 'othercontacts'
  ARG_OU = 'ou'
  ARG_OUS = 'ous'
  ARG_OUSHAREDDRIVE = 'oushareddrive'
  ARG_OUSHAREDDRIVES = 'oushareddrives'
  ARG_OUTOFOFFICE = 'outofoffice'
  ARG_OUTOFOFFICES = 'outofoffices'
  ARG_OUTREE = 'outree'
  ARG_OWNERSHIP = 'ownership'
  ARG_PARTICIPANTS = 'participants'
  ARG_PEOPLE = 'people'
  ARG_PEOPLECONTACT = 'peoplecontact'
  ARG_PEOPLECONTACTS = 'peoplecontacts'
  ARG_PEOPLECONTACTGROUP = 'peoplecontactgroup'
  ARG_PEOPLECONTACTGROUPS = 'peoplecontactgroups'
  ARG_PEOPLECONTACTPHOTO = 'peoplecontactphoto'
  ARG_PEOPLECONTACTPHOTOS = 'peoplecontactphotos'
  ARG_PEOPLEOTHERCONTACT = 'peopleothercontact'
  ARG_PEOPLEOTHERCONTACTS = 'peopleothercontacts'
  ARG_PEOPLEPROFILE = 'peopleprofile'
  ARG_PEOPLEPROFILES = 'peopleprofiles'
  ARG_PERMISSION = 'permission'
  ARG_PERMISSIONS = 'permissions'
  ARG_PHOTO = 'photo'
  ARG_POP = 'pop'
  ARG_POP3 = 'pop3'
  ARG_PRINTER = 'printer'
  ARG_PRINTERS = 'printers'
  ARG_PRINTERMODEL = 'printermodel'
  ARG_PRINTERMODELS = 'printermodels'
  ARG_PRIVILEGES = 'privileges'
  ARG_PROFILE = 'profile'
  ARG_PROFILE_PHOTO = 'profilephoto'
  ARG_PROJECT = 'project'
  ARG_PROJECTS = 'projects'
  ARG_RESELLERCUSTOMER = 'resellercustomer'
  ARG_RESELLERCUSTOMERS = 'resellercustomers'
  ARG_RESELLERSUBSCRIPTION = 'resellersubscription'
  ARG_RESELLERSUBSCRIPTIONS = 'resellersubscriptions'
  ARG_RESETPIV = 'resetpiv'
  ARG_RESOLDCUSTOMER = 'resoldcustomer'
  ARG_RESOLDCUSTOMERS = 'resoldcustomers'
  ARG_RESOLDSUBSCRIPTION = 'resoldsubscription'
  ARG_RESOLDSUBSCRIPTIONS = 'resoldsubscriptions'
  ARG_RESOURCE = 'resource'
  ARG_RESOURCES = 'resources'
  ARG_ROLE = 'role'
  ARG_ROLES = 'roles'
  ARG_SAKEY = 'sakey'
  ARG_SAKEYS = 'sakeys'
  ARG_SCHEMA = 'schema'
  ARG_SCHEMAS = 'schemas'
  ARG_SECCALS = 'seccals'
  ARG_SENDAS = 'sendas'
  ARG_SERVICEACCOUNT = 'serviceaccount'
  ARG_SETTINGS = 'settings'
  ARG_SHAREDDRIVE = 'shareddrive'
  ARG_SHAREDDRIVES = 'shareddrives'
  ARG_SHAREDDRIVEACLS = 'shareddriveacls'
  ARG_SHAREDDRIVEINFO = 'shareddriveinfo'
  ARG_SHAREDDRIVEORGANIZERS = 'shareddriveorganizers'
  ARG_SHAREDDRIVETHEMES = 'shareddrivethemes'
  ARG_SHEET = 'sheet'
  ARG_SHEETS = 'sheets'
  ARG_SHEETRANGE = 'sheetrange'
  ARG_SHEETRANGES = 'sheetranges'
  ARG_SIG = 'sig'
  ARG_SIGNATURE = 'signature'
  ARG_SIGNJWTSERVICEACCOUNT = 'signjwtserviceaccount'
  ARG_SITE = 'site'
  ARG_SITES = 'sites'
  ARG_SITEACL = 'siteacl'
  ARG_SITEACLS = 'siteacls'
  ARG_SITEACTIVITY = 'siteactivity'
  ARG_SMIME = 'smime'
  ARG_SMIMES = 'smimes'
  ARG_STORAGEBUCKET = 'storagebucket'
  ARG_STORAGEBUCKETS = 'storagebuckets'
  ARG_STORAGEFILE = 'storagefile'
  ARG_STORAGEFILES = 'storagefiles'
  ARG_SVCACCT = 'svcacct'
  ARG_SVCACCTS = 'svcaccts'
  ARG_TASK = 'task'
  ARG_TASKS = 'tasks'
  ARG_TASKLIST = 'tasklist'
  ARG_TASKLISTS = 'tasklists'
  ARG_TEAMDRIVE = 'teamdrive'
  ARG_TEAMDRIVES = 'teamdrives'
  ARG_TEAMDRIVEACLS = 'teamdriveacls'
  ARG_TEAMDRIVEINFO = 'teamdriveinfo'
  ARG_TEAMDRIVEORGANIZERS = 'teamdriveorganizers'
  ARG_TEAMDRIVETHEMES = 'teamdrivethemes'
  ARG_THREAD = 'thread'
  ARG_THREADS = 'threads'
  ARG_TOKEN = 'token'
  ARG_TOKENS = 'tokens'
  ARG_TRANSFER = 'transfer'
  ARG_TRANSFERS = 'transfers'
  ARG_TRANSFERAPPS = 'transferapps'
  ARG_TRUSTEDAPPS = 'trustedapps'
  ARG_USER = 'user'
  ARG_USERS = 'users'
  ARG_USERCOUNTSBYORGUNIT = 'usercountsbyorgunit'
  ARG_USERINVITATION = 'userinvitation'
  ARG_USERINVITATIONS = 'userinvitations'
  ARG_USERLIST = 'userlist'
  ARG_VACATION = 'vacation'
  ARG_VAULTCOUNT = 'vaultcount'
  ARG_VAULTCOUNTS = 'vaultcounts'
  ARG_VAULTEXPORT = 'vaultexport'
  ARG_VAULTEXPORTS = 'vaultexports'
  ARG_VAULTHOLD = 'vaulthold'
  ARG_VAULTHOLDS = 'vaultholds'
  ARG_VAULTMATTER = 'vaultmatter'
  ARG_VAULTMATTERS = 'vaultmatters'
  ARG_VAULTQUERY = 'vaultquery'
  ARG_VAULTQUERIES = 'vaultqueries'
  ARG_VERIFICATION = 'verification'
  ARG_VERIFICATIONCODES = 'verificationcodes'
  ARG_VERIFY = 'verify'
  ARG_WORKINGLOCATION = 'workinglocation'
  ARG_WORKINGLOCATIONS = 'workinglocations'
  ARG_YOUTUBECHANNEL = 'youtubechannel'
  ARG_YOUTUBECHANNELS = 'youtubechannels'
# Lists of arguments for use in checkArgumentPresent
  CLEAR_NONE_ARGUMENT = ['clear', 'none',]

# Object BNF names
  OB_ACCESS_LEVEL_NAME = 'AccessLevelName'
  OB_ACCESS_TOKEN = 'AccessToken'
  OB_ACL_SCOPE = 'ACLScope'
  OB_ACL_SCOPE_ENTITY = 'ACLScopeEntity'
  OB_ALERT_ID = 'AlertID'
  OB_ALIAS_ENTITY = 'AliasEntity'
  OB_API_SCOPE_URL_LIST = 'APIScopeURLList'
  OB_APP_ID = 'AppID'
  OB_ARGUMENT = 'argument'
  OB_ASP_ID_LIST = 'ASPIDList'
  OB_ASSET_ID = 'AssetID'
  OB_BROWSER_ENROLLEMNT_TOKEN_ID = 'BrowserEnrollmentTokenID'
  OB_BROWSER_ENTITY = 'BrowserEntity'
  OB_BUILDING_ID = 'BuildingID'
  OB_CALENDAR_ENTITY = 'CalendarEntity'
  OB_CALENDAR_ITEM = 'CalendarItem'
  OB_CHANNEL_CUSTOMER_ID = 'ChannelCustomerID'
  OB_CHARACTER = 'Character'
  OB_CHAR_SET = 'CharacterSet'
  OG_CHAT_ATTACHMENT = 'ChatAttachment'
  OB_CHAT_EVENT = 'ChatEvent'
  OB_CHAT_MEMBER = 'ChatMember'
  OB_CHAT_MESSAGE = 'ChatMessage'
  OB_CHAT_MESSAGE_ID = 'ChatMessageID'
  OB_CHAT_SPACE = 'ChatSpace'
  OB_CHAT_SPACE_LIST = 'ChatSpaceList'
  OB_CHAT_THREAD = 'ChatThread'
  OB_CHROMEPROFILE_ID = 'ChromeProfileId'
  OB_CHROME_VERSION = 'ChromeVersion'
  OB_CIDR_NETMASK = 'CIDRnetmask'
  OB_CIGROUP_ALIAS_LIST = "CIGroupAliasList"
  OB_CIPOLICY_NAME_ENTITY = 'CIPolicyNameEntity'
  OB_CLASSIFICATION_LABEL_ID = 'ClassificationLabelID'
  OB_CLASSIFICATION_LABEL_NAME = 'ClassificationLabelName'
  OB_CLASSIFICATION_LABEL_PERMISSION_NAME = 'ClassificationLabelPermissionName'
  OB_CLASSIFICATION_LABEL_FIELD_ID = 'ClassificationLabelFieldID'
  OB_CLASSIFICATION_LABEL_SELECTION_ID_LIST = 'ClassificationLabelSelectionIDList'
  OB_CLASSROOM_INVITATION_ID_ENTITY = 'ClassroomInvitationIDEntity'
  OB_CLIENT_ID = 'ClientID'
  OB_COLLABORATOR_ITEM = 'CollaboratorItem'
  OB_COLLABORATOR_ENTITY = 'CollaboratorEntity'
  OB_COMMAND_ID = 'CommandID'
  OB_CONTACT_EMAIL_TYPE = 'ContactEmailType'
  OB_CONTACT_ENTITY = 'ContactEntity'
  OB_CONTACT_GROUP_ENTITY = 'ContactGroupEntity'
  OB_CONTACT_GROUP_ITEM = 'ContactGroupItem'
  OB_COURSE_ALIAS = 'CourseAlias'
  OB_COURSE_ALIAS_ENTITY = 'CourseAliasEntity'
  OB_COURSE_ANNOUNCEMENT_ID_ENTITY = "CourseAnnouncementIDEntity"
  OB_COURSE_ANNOUNCEMENT_STATE_LIST = "CourseAnnouncementStateList"
  OB_COURSE_ENTITY = 'CourseEntity'
  OB_COURSE_ID = 'CourseID'
  OB_COURSE_MATERIAL_ID_ENTITY = 'CourseMaterialIDEntity'
  OB_COURSE_MATERIAL_STATE_LIST = "CourseMaterialStateList"
  OB_COURSE_STATE_LIST = "CourseStateList"
  OB_COURSE_SUBMISSION_ID_ENTITY = "CourseSubmissionIDEntity"
  OB_COURSE_SUBMISSION_STATE_LIST = "CourseSubmissionStateList"
  OB_COURSE_TOPIC = 'CourseTopic'
  OB_COURSE_TOPIC_ENTITY = "CourseTopicEntity"
  OB_COURSE_TOPIC_ID = 'CourseTopicID'
  OB_COURSE_TOPIC_ID_ENTITY = "CourseTopicIDEntity"
  OB_COURSE_WORK_ID_ENTITY = 'CourseWorkIDEntity'
  OB_COURSE_WORK_STATE_LIST = "CourseWorkStateList"
  OB_CROS_DEVICE_ENTITY = 'CrOSDeviceEntity'
  OB_CROS_ENTITY = 'CrOSEntity'
  OB_CSE_KEYPAIR_ID = 'CSEKeyPairID'
  OB_CUSTOMER_ID = 'CustomerID'
  OB_CUSTOMER_AUTH_TOKEN = 'CustomerAuthToken'
  OB_DEVICE_FILE_ENTITY = 'DeviceFileEntity'
  OB_DEVICE_ENTITY = 'DeviceEntity'
  OB_DEVICE_ID = 'DeviceID'
  OB_DEVICE_ID_LIST = 'DeviceIDList'
  OB_DEVICE_USER_ENTITY = 'DeviceUserEntity'
  OB_DEVICE_USER_ID = 'DeviceUserID'
  OB_DOMAIN_ALIAS = 'DomainAlias'
  OB_DOMAIN_NAME = 'DomainName'
  OB_DOMAIN_NAME_ENTITY = 'DomainNameEntity'
  OB_DOMAIN_NAME_LIST = 'DomainNameList'
  OB_DRIVE_FILE_ENTITY = 'DriveFileEntity'
  OB_DRIVE_FILE_ID = 'DriveFileID'
  OB_DRIVE_FILE_NAME = 'DriveFileName'
  OB_DRIVE_FILE_PERMISSION_ENTITY = 'DriveFilePermissionEntity'
  OB_DRIVE_FILE_PERMISSION_ID = 'DriveFilePermissionID'
  OB_DRIVE_FILE_PERMISSION_ID_ENTITY = 'DriveFilePermissionIDEntity'
  OB_DRIVE_FILE_REVISION_ID = 'DriveFileRevisionID'
  OB_DRIVE_FOLDER_ID = 'DriveFolderID'
  OB_DRIVE_FOLDER_ID_LIST = 'DriveFolderIDList'
  OB_DRIVE_FOLDER_NAME = 'DriveFolderName'
  OB_DRIVE_FOLDER_NAME_LIST = 'DriveFolderNameList'
  OB_DRIVE_FOLDER_PATH = 'DriveFolderPath'
  OB_EMAIL_ADDRESS = 'EmailAddress'
  OB_EMAIL_ADDRESS_ENTITY = 'EmailAddressEntity'
  OB_EMAIL_ADDRESS_LIST = 'EmailAddressList'
  OB_EMAIL_ADDRESS_OR_UID = 'EmailAaddress|UniqueID'
  OB_EMAIL_REPLACEMENT = 'EmailReplacement'
  OB_ENTITY = 'Entity'
  OB_ENTITY_TYPE = 'EntityType'
  OB_EVENT_ID = 'EventID'
  OB_EVENT_ID_ENTITY = 'EventIDEntity'
  OB_EVENT_NAME_LIST = "EventNameList"
  OB_EXPORT_ITEM = 'ExportItem'
  OB_FIELD_NAME = 'FieldName'
  OB_FIELD_NAME_LIST = "FieldNameList"
  OB_FIELDS = 'Fields'
  OB_FILE_NAME = 'FileName'
  OB_FILE_NAME_FIELD_NAME = OB_FILE_NAME+'(:'+OB_FIELD_NAME+')+'
  OB_FILE_NAME_OR_URL = 'FileName|URL'
  OB_FILE_PATH = 'FilePath'
  OB_FILTER_ID_ENTITY = 'FilterIDEntity'
  OB_FORMAT_LIST = 'FormatList'
  OB_GAM_ARGUMENT_LIST = 'GAM argument list'
  OB_GROUP_ENTITY = 'GroupEntity'
  OB_GROUP_ITEM = 'GroupItem'
  OB_GROUP_ROLE_LIST = 'GroupRoleList'
  OB_GROUP_TYPE_LIST = 'GroupTypeList'
  OB_GUARDIAN_INVITATION_ID = 'GuardianInvitationID'
  OB_GUARDIAN_INVITATION_ID_ENTITY = 'GuardianInvitationIDEntity'
  OB_GUARDIAN_ENTITY = 'GuardianEntity'
  OB_GUARDIAN_ITEM = 'GuardianItem'
  OB_GUARDIAN_STATE_LIST = 'GuardianStateList'
  OB_HOLD_ITEM = 'HoldItem'
  OB_HOST_NAME = 'HostName'
  OB_ICALUID = 'iCalUID'
  OB_ID_TOKEN = 'IDToken'
  OB_JOB_ID = 'JobID'
  OB_JSON_DATA = 'JSONData'
  OB_LABEL_COLOR_HEX = 'LabelColorHex'
  OB_LABEL_ID = 'LabelID'
  OB_LABEL_ID_LIST = 'LabelIDLIst'
  OB_LABEL_NAME = 'LabelName'
  OB_LABEL_NAME_LIST = 'LabelNameList'
  OB_LANGUAGE_LIST = 'LanguageList'
  OB_LOOKERSTUDIO_PERMISSION_ENTITY = 'LookerStudioPermissionEntity'
  OB_MATTER_ITEM = 'MatterItem'
  OB_MATTER_ITEM_LIST = 'MatterItemList'
  OB_MEET_CONFERENCE_NAME = 'MeetConferenceName'
  OB_MESSAGE_ID = 'MessageID'
  OB_MIMETYPE = 'MimeType'
  OB_MIMETYPE_LIST = 'MimeTypeList'
  OB_MOBILE_DEVICE_ENTITY = 'MobileDeviceEntity'
  OB_MOBILE_ENTITY = 'MobileEntity'
  OB_NETWORK_ID = 'networkID'
  OB_NAME = 'Name'
  OB_ORGANIZER_TYPE_LIST = 'OrganizerTypeList'
  OB_ORGUNIT_ENTITY = 'OrgUnitEntity'
  OB_ORGUNIT_ITEM = 'OrgUnitItem'
  OB_ORGUNIT_PATH = 'OrgUnitPath'
  OB_PARAMETER_VALUE = 'ParameterValue'
  OB_PASSWORD = 'Password'
  OB_PERMISSION_ID_LIST = 'PermissionIDList'
  OB_PERMISSION_ROLE_LIST = 'PermissionRoleList'
  OB_PERMISSION_TYPE_LIST = 'PermissionTypeList'
  OB_PHOTO_FILENAME_PATTERN = 'FilenameNamePattern'
  OB_PRINTER_ID = 'PrinterID'
  OB_PRIVILEGE_LIST = 'PrivilegeList'
  OB_PRODUCT_ID = 'ProductID'
  OB_PRODUCT_ID_LIST = 'ProductIDList'
  OB_PROJECT_ID = 'ProjectID'
  OB_PROJECT_ID_ENTITY = 'ProjectIDEntity'
  OB_PROPERTY_KEY = 'PropertyKey'
  OB_PROPERTY_VALUE = 'PropertyValue'
  OB_QUERY = 'Query'
  OB_QUERY_ITEM = 'QueryItem'
  OB_QUERY_LIST = 'QueryList'
  OB_RECURRENCE = 'RRULE EXRULE RDATE and EXDATE lines'
  OB_REQUEST_ID = 'RequestID'
  OB_RESELLER_ID = 'ResellerID'
  OB_RESOURCE_ENTITY = 'ResourceEntity'
  OB_RESOURCE_ID = 'ResourceID'
  OB_RE_PATTERN = 'REPattern'
  OB_RE_SUBSTITUTION = 'RESubstitution'
  OB_ROLE_ASSIGNMENT_ID = 'RoleAssignmentID'
  OB_ROLE_ITEM = 'RoleItem'
  OB_ROLE_LIST = 'RoleList'
  OB_SCHEMA_ENTITY = 'SchemaEntity'
  OB_SCHEMA_NAME = 'SchemaName'
  OB_SCHEMA_NAME_FIELD_NAME = 'SchemaName.FieldName'
  OB_SCHEMA_NAME_LIST = 'SchemaNameList'
  OB_SECTION_NAME = 'SectionName'
  OB_SERIAL_NUMBER = 'SerialNumber'
  OB_SERIAL_NUMBER_LIST = 'SerialNumberList'
  OB_SERVICE_NAME_LIST = 'ServiceNameList'
  OB_SERVICE_ACCOUNT_KEY_LIST = 'ServiceAccountKeyList'
  OB_SHAREDDRIVE_ID = 'SharedDriveID'
  OB_SHAREDDRIVE_ID_LIST = 'SharedDriveIDList'
  OB_SHAREDDRIVE_NAME = 'SharedDriveName'
  OB_SHEET_ENTITY = 'SheetEntity'
  OB_SITE_ENTITY = 'SiteEntity'
  OB_SKU_ID = 'SKUID'
  OB_SKU_ID_LIST = 'SKUIDList'
  OB_SMIME_ID = 'S/MIMEID'
  OB_SMTP_HOST_NAME = 'SMTPHostName'
  OB_SPREADSHEET_JSON_CREATEREQUEST = 'SpreadsheetJSONCreateRequest'
  OB_SPREADSHEET_JSON_RANGEVALUESLIST = 'SpreadsheetJSONRangeValuesList'
  OB_SPREADSHEET_JSON_UPDATEREQUEST = 'SpreadsheetJSONUpdateRequest'
  OB_SPREADSHEET_JSON_VALUES = 'SpreadsheetJSONValues'
  OB_SPREADSHEET_RANGE = 'SpreadsheetRange'
  OB_SPREADSHEET_RANGE_LIST = 'SpreadsheetRangeList'
  OB_STATE_NAME_LIST = "StateNameList"
  OB_STRING = 'String'
  OB_STRING_LIST = 'StringList'
  OB_STUDENT_ITEM = 'StudentItem'
  OB_TAG = 'Tag'
  OB_TASK_ID = 'TaskID'
  OB_TASKLIST_ID = 'TaskListID'
  OB_TASKLIST_ID_ENTITY = 'TaskListIDEntity'
  OB_TASKLIST_ID_TASK_ID = 'TaskListIDTaskID'
  OB_TASKLIST_ID_TASK_ID_ENTITY = 'TaskListIDTaskIDEntity'
  OB_THREAD_ID = 'ThreadID'
  OB_TIME_LIST = 'TimeList'
  OB_TRANSFER_ID = 'TransferID'
  OB_URI = 'URI'
  OB_URL = 'URL'
  OB_URL_LIST = 'URLList'
  OB_USER_ENTITY = 'UserEntity'
  OB_USER_ITEM = 'UserItem'
  OB_USER_NAME = 'UserName'
  OB_YOUTUBE_CHANNEL_ID_LIST = 'YouTubeChannelIDlist'

#
# Error message types; keys into ARGUMENT_ERROR_NAMES; arbitrary values but must be unique
  ARGUMENT_BLANK = 'blnk'
  ARGUMENT_DEPRECATED = 'depr'
  ARGUMENT_EMPTY = 'empt'
  ARGUMENT_EXTRANEOUS = 'extr'
  ARGUMENT_INVALID = 'inva'
  ARGUMENT_INVALID_CHOICE = 'invc'
  ARGUMENT_MISSING = 'miss'
# ARGUMENT_ERROR_NAMES[0] is plural,ARGUMENT_ERROR_NAMES[1] is singular
# These values can be translated into other languages
  ARGUMENT_ERROR_NAMES = {
    ARGUMENT_BLANK: ['Blank arguments', 'Blank argument'],
    ARGUMENT_DEPRECATED: ['Deprecated arguments', 'Deprecated argument'],
    ARGUMENT_EMPTY: ['Empty arguments', 'Empty argument'],
    ARGUMENT_EXTRANEOUS: ['Extra arguments', 'Extra argument'],
    ARGUMENT_INVALID: ['Invalid arguments', 'Invalid argument'],
    ARGUMENT_INVALID_CHOICE: ['Invalid choices', 'Invalid choice ({0})'],
    ARGUMENT_MISSING: ['Missing arguments', 'Missing argument'],
    }

  def __init__(self):
    self.argv = []
    self.argvI = 0
    self.argvLen = 0
    self.argvIsave = 0
    self.origArgv = []
    self.origArgvI = 0
    self.origArgvLen = 0
    self.encoding = 'utf-8'

# Initialize arguments
  def InitializeArguments(self, args):
    self.argv = args[:]
    self.argvI = 1
    self.argvLen = len(self.argv)

# Number of arguments remaining
  def NumArgumentsRemaining(self):
    return self.argvLen - self.argvI

# Any arguments remaining
  def ArgumentsRemaining(self):
    return self.argvI < self.argvLen

# Multiple arguments remaining
  def MultipleArgumentsRemaining(self):
    return not self.argvI+1 == self.argvLen

# All arguments
  def AllArguments(self):
    return self.argv

# Specific argument
  def Argument(self, index):
    return self.argv[index]

# Current argument
  def Current(self):
    return self.argv[self.argvI]

# Previous argument
  def Previous(self):
    return self.argv[self.argvI-1]

# Remaining arguments
  def Remaining(self):
    return self.argv[self.argvI:]

# Argument location
  def Location(self):
    return self.argvI

# Advance to next argument
  def Advance(self):
    self.argvI += 1

# Backup to previous argument
  def Backup(self):
    self.argvI -= 1

# Save argument location
  def SaveLocation(self):
    self.argvIsave = self.argvI

# Reset argument location
  def ResetLocation(self, offset):
    self.argvI = self.argvIsave+offset

# Set argument location
  def SetLocation(self, location):
    self.argvI = location

# Set encoding
  def SetEncoding(self, encoding):
    self.encoding = encoding

# Concatenate list members, any item containing spaces, commas or single quotes is enclosed in ""
  @staticmethod
  def QuotedArgumentList(items):
    return ' '.join([item if item and (item.find(' ') == -1) and (item.find(',') == -1) and (item.find("'") == -1) else '"'+item+'"' for item in items])

# Mark bad argument in command line
  def CommandLineWithBadArgumentMarked(self, extraneous):
    if extraneous:
      return f'Command: {self.QuotedArgumentList(self.argv[:self.argvI])} >>>{self.QuotedArgumentList(self.argv[self.argvI:])}<<<\n'
    if self.ArgumentsRemaining():
      return f'Command: {self.QuotedArgumentList(self.argv[:self.argvI])} >>>{self.QuotedArgumentList([self.argv[self.argvI]])}<<< {self.QuotedArgumentList(self.argv[self.argvI+1:])}\n'
    return f'Command: {self.QuotedArgumentList(self.argv)} >>><<<\n'

# Deprecated command
  def CommandDeprecated(self):
    return f'{self.QuotedArgumentList(self.argv)}\n'

# Peek to see if next argument is in choices
  def PeekArgumentPresent(self, choices):
    if self.ArgumentsRemaining():
      choiceList = choices if isinstance(choices, (list, set)) else [choices]
      choice = self.Current().strip().lower().replace('_', '')
      if choice and choice in choiceList:
        return True
    return False

# Look ahead to see if argument is present
  def ArgumentIsAhead(self, argument):
    self.SaveLocation()
    while self.ArgumentsRemaining():
      if argument == self.Current().strip().lower().replace('_', ''):
        self.ResetLocation(0)
        return True
      self.Advance()
    self.ResetLocation(0)
    return False

# Merge new arguments into current argument list
  def MergeArguments(self, arguments):
    self.origArgv = self.argv[:]
    self.origArgvI = self.argvI
    self.origArgvLen = self.argvLen
    self.argv = self.argv[0:self.argvI]+arguments+self.argv[self.argvI:]
    self.argvLen += len(arguments)

# Restore
  def RestoreArguments(self):
    self.argv = self.origArgv[:]
    self.argvI = self.origArgvI
    self.argvLen = self.origArgvLen



# -*- coding: utf-8 -*-

# Copyright (C) 2023 Ross Scroggs All Rights Reserved.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""GAM indent processing

"""

class GamIndent():

  INDENT_SPACES_PER_LEVEL = '  '

  def __init__(self):
    self.indent = 0

  def Reset(self):
    self.indent = 0

  def Increment(self):
    self.indent += 1

  def Decrement(self):
    self.indent -= 1

  def Spaces(self):
    return self.INDENT_SPACES_PER_LEVEL*self.indent

  def SpacesSub1(self):
    return self.INDENT_SPACES_PER_LEVEL*(self.indent-1)

  def MultiLineText(self, message, n=0):
    return message.replace('\n', f'\n{self.INDENT_SPACES_PER_LEVEL*(self.indent+n)}').rstrip()



#
# Copyright (C) 2008 Google Inc.
#
# Licensed under the Apache License 2.0;



# __author__ = 'api.jscudder (Jeff Scudder)'

import urllib.error
import urllib.parse
import urllib.parse
import urllib.request

DEFAULT_PROTOCOL = 'http'
DEFAULT_PORT = 80


def parse_url(url_string):
    """Creates a Url object which corresponds to the URL string.

    This method can accept partial URLs, but it will leave missing
    members of the Url unset.
    """
    parts = urllib.parse.urlparse(url_string)
    url = Url()
    if parts[0]:
        url.protocol = parts[0]
    if parts[1]:
        host_parts = parts[1].split(':')
        if host_parts[0]:
            url.host = host_parts[0]
        if len(host_parts) > 1:
            url.port = host_parts[1]
    if parts[2]:
        url.path = parts[2]
    if parts[4]:
        param_pairs = parts[4].split('&')
        for pair in param_pairs:
            pair_parts = pair.split('=')
            if len(pair_parts) > 1:
                url.params[urllib.parse.unquote_plus(pair_parts[0])] = (
                    urllib.parse.unquote_plus(pair_parts[1]))
            elif len(pair_parts) == 1:
                url.params[urllib.parse.unquote_plus(pair_parts[0])] = None
    return url


class Url(object):
    """Represents a URL and implements comparison logic.

    URL strings which are not identical can still be equivalent, so this object
    provides a better interface for comparing and manipulating URLs than
    strings. URL parameters are represented as a dictionary of strings, and
    defaults are used for the protocol (http) and port (80) if not provided.
    """

    def __init__(self, protocol=None, host=None, port=None, path=None,
                 params=None):
        self.protocol = protocol
        self.host = host
        self.port = port
        self.path = path
        self.params = params or {}

    def to_string(self):
        url_parts = ['', '', '', '', '', '']
        if self.protocol:
            url_parts[0] = self.protocol
        if self.host:
            if self.port:
                url_parts[1] = ':'.join((self.host, str(self.port)))
            else:
                url_parts[1] = self.host
        if self.path:
            url_parts[2] = self.path
        if self.params:
            url_parts[4] = self.get_param_string()
        return urllib.parse.urlunparse(url_parts)

    def get_param_string(self):
        param_pairs = []
        for key, value in self.params.items():
            param_pairs.append('='.join((urllib.parse.quote_plus(key),
                                         urllib.parse.quote_plus(str(value)))))
        return '&'.join(param_pairs)

    def get_request_uri(self):
        """Returns the path with the parameters escaped and appended."""
        param_string = self.get_param_string()
        if param_string:
            return '?'.join([self.path, param_string])
        else:
            return self.path

    def __cmp__(self, other):
        if not isinstance(other, Url):
            return cmp(self.to_string(), str(other))
        difference = 0
        # Compare the protocol
        if self.protocol and other.protocol:
            difference = cmp(self.protocol, other.protocol)
        elif self.protocol and not other.protocol:
            difference = cmp(self.protocol, DEFAULT_PROTOCOL)
        elif not self.protocol and other.protocol:
            difference = cmp(DEFAULT_PROTOCOL, other.protocol)
        if difference != 0:
            return difference
        # Compare the host
        difference = cmp(self.host, other.host)
        if difference != 0:
            return difference
        # Compare the port
        if self.port and other.port:
            difference = cmp(self.port, other.port)
        elif self.port and not other.port:
            difference = cmp(self.port, DEFAULT_PORT)
        elif not self.port and other.port:
            difference = cmp(DEFAULT_PORT, other.port)
        if difference != 0:
            return difference
        # Compare the path
        difference = cmp(self.path, other.path)
        if difference != 0:
            return difference
        # Compare the parameters
        return cmp(self.params, other.params)

    def __str__(self):
        return self.to_string()



# -*- coding: utf-8 -*-

# Copyright (C) 2025 Ross Scroggs All Rights Reserved.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""GAM messages

"""

# These values can be translated into other languages
# Project creation messages in order of appearance
CREATING_PROJECT = 'Creating project "{0}"...\n'
CHECK_INTERRUPTED = 'Check interrupted'
CHECKING_PROJECT_CREATION_STATUS = 'Checking project creation status...\n'
NO_RIGHTS_GOOGLE_CLOUD_ORGANIZATION = 'Looks like you have no rights to your Google Cloud Organization.\nAttempting to fix that...\n'
YOUR_ORGANIZATION_NAME_IS = 'Your organization name is {0}\n'
YOU_HAVE_NO_RIGHTS_TO_CREATE_PROJECTS_AND_YOU_ARE_NOT_A_SUPER_ADMIN = 'You have no rights to create projects for your organization and you don\'t seem to be a super admin! Sorry, there\'s nothing more I can do.'
LOOKS_LIKE_NO_ONE_HAS_RIGHTS_TO_YOUR_GOOGLE_CLOUD_ORGANIZATION_ATTEMPTING_TO_GIVE_YOU_CREATE_RIGHTS = 'Looks like no one has rights to your Google Cloud Organization. Attempting to give you create rights...\n'
THE_FOLLOWING_RIGHTS_SEEM_TO_EXIST = 'The following rights seem to exist:\n'
GIVING_LOGIN_HINT_THE_CREATOR_ROLE = 'Giving {0} the role of {1}...\n'
ACCEPT_CLOUD_TOS = '''
Please go to:

https://console.cloud.google.com/projectselector2/home/dashboard?supportedpurview=project

sign in as {0} and accept the Terms of Service (ToS). As soon as you've accepted the ToS popup, you can return here and press enter.\n'''

PROJECT_STILL_BEING_CREATED_SLEEPING = 'Project still being created. Sleeping {0} seconds\n'
FAILED_TO_CREATE_PROJECT = 'Failed to create project: {0}\n'
SETTING_GAM_PROJECT_CONSENT_SCREEN_CREATING_CLIENT = 'Setting GAM project consent screen, creating client...\n'
CREATE_CLIENT_INSTRUCTIONS = '''
Please go to:

  {0}

 1. If "+ CREATE CLIENT" is on the screen, skip to step 14
 2. Click "GET STARTED"
 3. Under "App Information", enter {1} or another value in "App name *"
 4. Under "App Information", enter {2} in "User support email *"
 5. Click "NEXT"
 6. Under "Audience", choose INTERNAL
 7. Click "NEXT"
 8. Under, "Contact Information", enter an email address in "Email addresses *"
 9. Click "NEXT"
10. Under "Finish", click "I agree to the Google API Services: User Data Policy."
11. Click "CONTINUE"
12. Click "CREATE"
13. Click "Clients" in the left-hand column
14. Click "+ CREATE CLIENT"
15. Choose "Desktop App" for "Application type"
16. Enter {1} or another value in "Name *"
17. Click "Create"
18. Under "Name", click your client name
19. Copy the "Client ID" value under "Additional information"
20. Paste it at the "Enter your Client ID: " prompt in your terminal
21. Press return/enter in your terminal
22. Switch back to the browser
23. Copy the "Client secret" value under "Client Secrets"
24. Paste it at the "Enter your Client Secret: " prompt in your terminal
25. Press return/enter in your terminal
26. Switch back to the browser
27. Click "OK"
28. These steps are complete
'''
ENTER_YOUR_CLIENT_ID = '\nEnter your Client ID: '
ENTER_YOUR_CLIENT_SECRET = '\nEnter your Client Secret: '
IS_NOT_A_VALID_CLIENT_ID = '''

{0}

Is not a valid Client ID.

Please make sure you are following the directions exactly and that there are no extra spaces in your Client ID.
'''
IS_NOT_A_VALID_CLIENT_SECRET = '''

{0}

Is not a valid Client Secret.

Please make sure you are following the directions exactly and that there are no extra spaces in your Client Secret.
'''
TRUST_GAM_CLIENT_ID = '''
It's important to mark the {0} Client ID as trusted by your Workspace instance.

Please go to:

  https://admin.google.com/ac/owl/list?tab=configuredApps

1. Click on: Configure new app
2. Enter the following Client ID value in Search for app:

  {1}

3. Press Search, select the {0} app, click
4. Keep the default scope or select a preferred scope that includes your GAM admin.
5. Press Continue
6. Select Trusted radio button, press Continue and Finish.
7. Press Confirm if Confirm parental consent pops up
8. Press enter here on the terminal once trust is complete.
'''

ENABLE_SERVICE_ACCOUNT_PRIVATE_KEY_UPLOAD = '''
Your workspace is configured to disable service account private key uploads.

Please go to:

  https://github.com/GAM-team/GAM/wiki/Authorization#authorize-service-account-key-uploads

Follow the steps to allow a service account private key upload for the project ({0}) just created.
Once those steps are completed, you can continue with your project authentication.
'''

YOUR_GAM_PROJECT_IS_CREATED_AND_READY_TO_USE = '''
That\'s it! Your GAM Project is created and ready to use.
Proceed to the authentication steps.
'''

# check|update service messages in order of appearance
SYSTEM_TIME_STATUS = 'System time status'
YOUR_SYSTEM_TIME_DIFFERS_FROM_GOOGLE = 'Your system time differs from {0} by {1}'
PRESS_ENTER_ONCE_AUTHORIZATION_IS_COMPLETE = 'Press enter once authorization is complete.'
SERVICE_ACCOUNT_API_DISABLED = '{0} not enabled. Please run "gam update project" and "gam user user@domain.com update serviceaccount"'
SERVICE_ACCOUNT_PRIVATE_KEY_AUTHENTICATION = 'Service Account Private Key Authentication'
SERVICE_ACCOUNT_CHECK_PRIVATE_KEY_AGE = 'Service Account Private Key age; Google recommends rotating keys on a routine basis'
SERVICE_ACCOUNT_PRIVATE_KEY_AGE = 'Service Account Private Key age: {0} days'
SERVICE_ACCOUNT_SKIPPING_KEY_AGE_CHECK = 'Skipping Private Key age check: {0} rotation not necessary'
UPDATE_PROJECT_TO_VIEW_MANAGE_SAKEYS = 'Please run "gam update project" to view/manage service account keys'
DOMAIN_WIDE_DELEGATION_AUTHENTICATION = 'Domain-wide Delegation authentication'
SCOPE_AUTHORIZATION_PASSED = '''All scopes PASSED!

Service Account Client name: {0} is fully authorized.
'''
SCOPE_AUTHORIZATION_UPDATE_PASSED = '''All scopes PASSED!
To authorize them (in case some scopes were unselected), please go to the following link in your browser:
{0}
    {1}

You will be directed to the Google Workspace admin console Security > API Controls > Domain-wide Delegation page
The "Add a new Client ID" box will open
Make sure that "Overwrite existing client ID" is checked
Click AUTHORIZE
When the box closes you're done
After authorizing it may take some time for this test to pass so wait a few moments and then try this command again.
'''
SCOPE_AUTHORIZATION_FAILED = '''Some scopes FAILED!
To authorize them, please go to the following link in your browser:
{0}
    {1}

You will be directed to the Google Workspace admin console Security > API Controls > Domain-wide Delegation page
The "Add a new Client ID" box will open
Make sure that "Overwrite existing client ID" is checked
Click AUTHORIZE
When the box closes you're done
After authorizing it may take some time for this test to pass so wait a few moments and then try this command again.
'''
# General messages
ACCESS_FORBIDDEN = 'Access Forbidden'
ACTION_APPLIED = 'Action Applied'
ACTION_IN_PROGRESS = 'Action {0} in progress'
ACTION_MAY_BE_DELAYED = 'Action may be delayed'
ADMIN_STATUS_CHANGED_TO = 'Admin Status Changed to'
ALL = 'All'
ALL_FOLDER_NAMES_MUST_BE_NON_BLANK = 'All folder names must be non-blank: {0}'
ALL_SKU_PRODUCTIDS_MUST_MATCH = 'All SKU productIds must match, {0} != {1}'
ALREADY_WAS_OWNER = 'Already was owner'
ALREADY_EXISTS = 'Already exists'
ALREADY_EXISTS_IN_TARGET_FOLDER = 'Already exists in {0}: {1}'
ALREADY_EXISTS_USE_MERGE_ARGUMENT = 'Already exists; use the "merge" argument to merge the labels'
API_ACCESS_DENIED = 'API access Denied'
API_CALLS_RETRY_DATA = 'API calls retry data\n'
API_CHECK_CLIENT_AUTHORIZATION = 'Please make sure the Client ID: {0} is authorized for the appropriate API or scopes:\n{1}\n\nRun: gam oauth create\n'
API_CHECK_SVCACCT_AUTHORIZATION = 'Please make sure the Service Account Client ID: {0} is authorized for the appropriate API or scopes:\n{1}\n\nRun: gam user {2} update serviceaccount\n'
API_ERROR_SETTINGS = 'API error, some settings not set'
ARE_BOTH_REQUIRED = 'Arguments {0} and {1} are both required'
ARE_MUTUALLY_EXCLUSIVE = 'Arguments {0} and {1} are mutually exclusive'
AS = 'as'
ATTENDEES_ADD = 'Add Attendees'
ATTENDEES_ADD_REMOVE = 'Add/Remove Attendees'
ATTENDEES_REMOVE = 'Remove Attendees'
AUTHORIZATION_FILE_ALREADY_EXISTS = '{0} already exists. Please delete or rename it before attempting to {1} project.'
AUTHENTICATION_FLOW_COMPLETE = '\nThe authentication flow has completed.'
AUTHENTICATION_FLOW_COMPLETE_CLOSE_BROWSER = 'The authentication flow has completed. You may close this browser window and return to {0}.'
AUTHENTICATION_FLOW_FAILED = 'The authentication flow failed, reissue command'
BAD_ENTITIES_IN_SOURCE = '{0} {1} {2} in source marked >>> <<< above'
BAD_REQUEST = 'Bad Request'
BATCH = 'Batch'
BATCH_CSV_LOOP_DASH_DEBUG_INCOMPATIBLE = '"gam {0} - ..." is not compatible with debugging. Disable debugging by setting debug_level = 0 in gam.cfg'
BATCH_CSV_PROCESSING_COMPLETE = '{0},0/{1},Processing complete\n'
BATCH_CSV_TERMINATE_N_PROCESSES = '{0},0/{1},Terminating {2} running {3}\n'
BATCH_CSV_WAIT_LIMIT = ', wait limit {0} seconds'
BATCH_CSV_WAIT_N_PROCESSES = '{0},0/{1},Waiting for {2} running {3} to finish before terminating{4}\n'
BATCH_NOT_PROCESSED_ERRORS = '{0}batch file: {1}, not processed, {2} {3}\n'
CALLING_GCLOUD_FOR_REAUTH = 'Calling gcloud for reauth credentials..."\n'
CAN_NOT_DELETE_USER_WITH_VAULT_HOLD = '{0}: The user may be (or have recently been) on Google Vault Hold and thus not eligible for deletion. You can check holds with "gam user {1} show vaultholds".'
CAN_NOT_BE_SPECIFIED_MORE_THAN_ONCE = 'Argument {0} can not be specified more than once'
CHAT_ADMIN_ACCESS_LIMITED_TO_ONE_USER = 'Chat adminaccess|asadmin limited to one user, {0} specified'
CHROME_TARGET_VERSION_FORMAT = r'^([a-z]+)-(\d+)$ or ^(\d{1,4}\.){1,4}$'
COLUMN_DOES_NOT_MATCH_ANY_INPUT_COLUMNS = '{0} column "{1}" does not match any input columns'
COLUMN_DOES_NOT_MATCH_ANY_OUTPUT_COLUMNS = '{0} column "{1}" does not match any output columns'
COMMAND_NOT_COMPATIBLE_WITH_ENABLE_DASA = 'gam {0} {1} is not compatible with enable_dasa = true in gam.cfg'
COMMIT_BATCH_COMPLETE = '{0},0/{1},commit-batch - running {2} finished, proceeding\n'
COMMIT_BATCH_WAIT_N_PROCESSES = '{0},0/{1},commit-batch - waiting for {2} running {3} to finish before proceeding\n'
CONFIRM_WIPE_YUBIKEY_PIV = 'This will wipe all YubiKey PIV keys and configuration from your YubiKey. Are you sure? (y/N) '
CONTACT_ADMINISTRATOR_FOR_PASSWORD = 'Contact administrator for password'
CONTACT_PHOTO_NOT_FOUND = 'Contact photo not found'
CONTAINS_AT_LEAST_1_ITEM = 'Contains at least 1 item'
COUNT_N_EXCEEDS_MAX_TO_PROCESS_M = 'Count {0} exceeds maximum to {1} {2}'
CORRUPT_FILE = 'Corrupt file'
COULD_NOT_FIND_ANY_YUBIKEY = 'Could not find any YubiKey\n'
COULD_NOT_FIND_YUBIKEY_WITH_SERIAL = 'Could not find YubiKey with serial number {0}\n'
CREATE_USER_NOTIFY_MESSAGE = 'Hello #givenname# #familyname#,\n\nYou have a new account at #domain#\nAccount details:\nUsername: #user#\nPassword: #password#\nStart using your new account by signing in at\nhttps://www.google.com/accounts/AccountChooser?Email=#user#&continue=https://workspace.google.com/dashboard\n'
CREATE_USER_NOTIFY_SUBJECT = 'Welcome to #domain#'
CSV_DATA_ALREADY_SAVED = 'CSV data already saved'
CSV_FILE_HEADERS = 'The CSV file ({0}) has the following headers:\n'
CSV_SAMPLE_COMMANDS = 'Here are the first {0} commands {1} will run\n'
DATA_FIELD_MISMATCH = 'datafield {0} does not match saved datafield {1}'
DATA_TRANSFER_COMPLETED = 'Data Transfer completed: {0}\n'
DATA_UPLOADED_TO_DRIVE_FILE = 'Data uploaded to Drive File'
DEFAULT_SMIME = 'Default S/MIME'
DELETED = 'Deleted'
DEVICE_LIST_BUG = 'GAM hit Google internal bug 237397223. Please file a Google Support ticket stating that you are encountering this bug.'
DEVICE_LIST_BUG_WORKAROUND_NOT_POSSIBLE = 'GAM workaround for this issue only works if orderby argument is not used and query does not contain \'register\'.'
DEVICE_LIST_BUG_ATTEMPTING_WORKAROUND = 'GAM is attempting to work around the bug by filtering for devices created on or after the newest we\'ve seen ({0})...\n'
DIRECTLY_IN_THE = ' directly in the {0}'
DISABLE_TLS_MIN_MAX = 'Execute: gam select default config tls_max_version "" tls_min_version "" save\n'
DISPLAYNAME_NOT_ALLOWED_WHEN_UPDATING_MULTIPLE_SCHEMAS = 'displayname not allowed when updating multiple schemas'
DOES_NOT_EXIST = 'Does not exist'
DOES_NOT_EXIST_OR_HAS_INVALID_FORMAT = '{0}: {1}, Does not exist or has invalid format, {2}'
DOES_NOT_MATCH = 'Does not match {0}'
DOMAIN_NOT_FOUND_IN_DNS = 'Domain not found in DNS!'
DOMAIN_NOT_VERIFIED_SECONDARY = 'Domain is not a verified secondary domain'
DONE_GENERATING_PRIVATE_KEY_AND_PUBLIC_CERTIFICATE = 'Done generating private key and public certificate'
DO_NOT_EXIST = 'Do not exist'
DOWNLOADING_AGAIN_AND_OVER_WRITING = 'Downloading again and over-writing...'
DUPLICATE = 'Duplicate'
DUPLICATE_ALREADY_A_ROLE = 'Duplicate, already a {0}'
DYNAMIC_GROUP_MEMBERSHIP_CANNOT_BE_MODIFIED = 'Dynamic group membership cannot be modified'
EITHER = 'Either'
EMAIL_ADDRESS_IS_UNMANAGED_ACCOUNT = 'The email address is an unmanaged account'
ENABLE_PROJECT_APIS_AUTOMATICALLY_OR_MANUALLY = 'Do you want to enable project APIs [a]utomatically or [m]anually? (a/m): '
ENTER_GSUITE_ADMIN_EMAIL_ADDRESS = '\nEnter your Google Workspace admin email address? '
ENTER_MANAGE_GCP_PROJECT_EMAIL_ADDRESS = '\nEnter your Google Workspace admin or GCP project manager email address authorized to manage project(s): {0}? '
ENTER_VERIFICATION_CODE_OR_URL = 'Enter verification code or paste "Unable to connect" URL from other computer (only URL data up to &scope required): '
ENTITY_DOES_NOT_EXIST = '{0} does not exist'
ENTITY_NAME_NOT_VALID = 'Entity Name Not Valid'
ERROR = 'error'
ERRORS = 'errors'
EVENT_IS_CANCELED = 'Event is canceled'
EXECUTE_GAM_OAUTH_CREATE = '\nPlease run\n\ngam oauth delete\ngam oauth create\n\n'
EXISTS = 'Exists'
EXPECTED = 'Expected'
EXPORT_NOT_COMPLETE = 'Export needs to be complete before downloading, current status is: {0}'
EXTRACTING_PUBLIC_CERTIFICATE = 'Extracting public certificate'
FAILED_PRECONDITION = 'Failed precondition'
FAILED_TO_PARSE_AS_JSON = 'Failed to parse as JSON'
FAILED_TO_PARSE_AS_LIST = 'Failed to parse as list'
FIELD_NOT_FOUND_IN_SCHEMA = 'Field {0} not found in schema {1}'
FILE_NOT_FOUND = 'File {0} not found'
FINISHED = 'Finished'
FILTER_CAN_ONLY_CONTAIN_ONE_CATEGORY_LABEL = 'Filter can only contain one CATEGORY label'
FILTER_CAN_ONLY_CONTAIN_ONE_USER_LABEL = 'Filter can only contain one USER label'
FOR = 'for'
FORBIDDEN = 'Forbidden'
FORMAT_NOT_AVAILABLE = 'Format ({0}) not available'
FORMAT_NOT_DOWNLOADABLE = 'Format not downloadable'
FROM = 'From'
FROM_LC = 'from'
FULL_PATH_MUST_START_WITH_DRIVE = 'fullpath must start with {0} or {1}'
GAM_BATCH_FILE_WRITTEN = 'GAM batch file {0} written\n'
GAM_LATEST_VERSION_NOT_AVAILABLE = 'GAM Latest Version information not available'
GAM_OUT_OF_MEMORY = 'GAM has run out of memory. If this is a large Google Workspace instance, you should use a 64-bit version of GAM on Windows or a 64-bit version of Python on other systems.'
GENERATING_NEW_PRIVATE_KEY = 'Generating new private key'
GETTING = 'Getting'
GETTING_ALL = 'Getting all'
GRANTING_RIGHTS_TO_ROTATE_ITS_OWN_PRIVATE_KEY = '{0} rights to rotate its own private key'
GOOGLE_DELEGATION_ERROR = 'Google delegation error, delegator and delegate both exist and are valid for delegation'
GOT = 'Got'
GROUP_MAPS_TO_MULTIPLE_OUS = 'File: {0}, Group: {1} references multiple OUs: {2}'
GROUP_MAPS_TO_OU_INVALID_ROW = 'File: {0}, Invalid row, must contain non-blank <EmailAddress> and <OrgUnitPath>: <{1}> <{2}>'
GUARDIAN_INVITATION_STATUS_NOT_PENDING = 'Guardian invitation status is not PENDING'
HAS_CHILD_ORGS = 'Has child {0}'
HAS_INVALID_FORMAT = '{0}: {1}, Has invalid format'
HEADER_NOT_FOUND_IN_CSV_HEADERS = 'Header "{0}" not found in CSV headers of "{1}".'
HELP_SYNTAX = 'Help: Syntax in file {0}\n'
HELP_WIKI = 'Help: Documentation is at {0}\n'
IGNORED = 'Ignored'
INSTRUCTIONS_CLIENT_SECRETS_JSON = 'Please run\n\ngam create|use project\ngam oauth create\n\nto create and authorize a Client account.\n'
INSTRUCTIONS_OAUTH2SERVICE_JSON = 'Please run\n\ngam create|use project\ngam user <user> update serviceaccount\n\nto create and authorize a Service account.\n'
INSUFFICIENT_PERMISSIONS_TO_PERFORM_TASK = 'Insufficient permissions to perform this task'
INTER_BATCH_WAIT_INCREASED = 'inter_batch_wait increased to {0:.2f}'
INVALID = 'Invalid'
INVALID_ALIAS = 'Invalid Alias'
INVALID_ATTENDEE_CHANGE = 'Invalid attendee change "{0}"'
INVALID_CHARSET = 'Invalid charset "{0}"'
INVALID_DATE_TIME_RANGE = '{0} {1} must be greater than/equal to {2} {3}'
INVALID_ENTITY = 'Invalid {0}, {1}'
INVALID_FILE_SELECTION_WITH_ADMIN_ACCESS = 'Invalid file selection with adminaccess|asadmin'
INVALID_GROUP = 'Invalid Group'
INVALID_HTTP_HEADER = 'Invalid http header data: {0}'
INVALID_JSON_INFORMATION = 'Google API reported Invalid JSON Information'
INVALID_JSON_SETTING = 'Invalid JSON setting'
INVALID_LIST = 'Invalid list'
INVALID_MEMBER = 'Invalid Member address'
INVALID_MESSAGE_ID = 'Invalid message id(s)'
INVALID_MIMETYPE = 'Invalid mimeType {0}, must be {1}'
INVALID_NUMBER_OF_CHAT_SPACE_MEMBERS = '{0} type {1} number of members, {2}, must be in range {3} to {4}'
INVALID_ORGUNIT = 'Invalid Organizational Unit'
INVALID_PATH = 'Invalid Path'
INVALID_PERMISSION_ATTRIBUTE_TYPE = 'permission attribute {0} not allowed with type {1}'
INVALID_REGION = 'See: https://github.com/GAM-team/GAM/wiki/Context-Aware-Access-Levels#caa-region-codes'
INVALID_QUERY = 'Invalid Query'
INVALID_RE = 'Invalid RE'
INVALID_REQUEST = 'Invalid Request'
INVALID_RESELLER_CUSTOMER_NAME = 'name must be: accounts/<ResellerID>/customers/<ChannelCustomerID>'
INVALID_ROLE = 'Invalid subkeyfield Role, must be one of: {0}'
INVALID_SCHEMA_VALUE = 'Invalid Schema Value'
INVALID_SCOPE = 'Invalid Scope'
INVALID_SITE = 'Invalid Site ({0}), must match pattern ({1})'
INVALID_TAG_SPECIFICATION = 'Invalid tag, expected field.subfield or field.subfield.subfield.string'
INVALID_TIMEOFDAY_RANGE = '{0} must be less than/equal to {1}'
IN_SKIPIDS = 'In skipids'
IN_THE = ' in the {0}'
IN_TRASH_AND_EXCLUDE_TRASHED = 'In Trash and excludeTrashed'
IS_EXPIRED_OR_REVOKED = '{0}: {1}, Is expired or has been revoked'
IS_NOT_DONE_CHECKING_IN_SECONDS = 'Is not done, checking again in {0} seconds'
IS_NOT_UNIQUE = 'Is not unique, {0}: {1}'
IS_REQD_TO_CHG_PWD_NO_DELEGATION = 'Is required to change password at next login. You must change password or clear changepassword flag for delegation.'
IS_SUSPENDED_NO_BACKUPCODES = 'User is suspended. You must unsuspend to process backupcodes'
IS_SUSPENDED_NO_DELEGATION = 'Is suspended. You must unsuspend for delegation.'
IS_YUBIKEY_INSERTED = 'Is YubiKey inserted?'
JSON_ERROR = 'JSON error "{0}" in file {1}'
JSON_KEY_NOT_FOUND = 'JSON key "{0}" not found in file {1}'
KIOSK_MODE_REQUIRED = ' This command ({0}) requires that the ChromeOS device be in Kiosk mode.'
LESS_THAN_1_SECOND = 'less than 1 second'
LIST_CHROMEOS_INVALID_INPUT_PAGE_TOKEN_RETRY = 'List ChromeOSdevices Invalid Input: pageToken retry'
LOGGING_INITIALIZATION_ERROR = 'Logging initialization error: {0}'
LOOKING_UP_GOOGLE_UNIQUE_ID = 'Looking up Google Unique ID'
MARKED_AS = 'Marked as'
MATCHED_THE_FOLLOWING = 'Matched the following'
MATTER_NOT_OPEN = 'Matter needs to be open, current state is: {0}'
MAXIMUM_OF = 'maximum of'
MEMBERSHIP_IS_PENDING_WILL_DELETE_ADD_TO_ACCEPT = 'Membership is pending, will delete and add to accept'
MIMETYPE_MISMATCH = 'Shortcut target mimeType {0} does not match actual target mimeType {1}'
MIMETYPE_NOT_PRESENT_IN_ATTACHMENT = 'MIME type not present in attachment'
MISMATCH_RE_SEARCH_REPLACE_SUBFIELDS = 'The subfield ({2}) in replace "{3}" exceeds the number of subfields ({0}) in search "{1}"'
MISMATCH_SEARCH_REPLACE_SUBFIELDS = 'The number of subfields ({0}) in search "{1}" does not match the number of subfields ({2}) in replace "{3}"'
MISSING_FIELDS = 'Missing fields: {0}\n'
MULTIPLE_BUILDINGS_SAME_NAME = '{0} {1} with the same (case-insensitive) name exist'
MULTIPLE_ENTITIES_FOUND = 'Multiple {0} ({1}) found, {2}'
MULTIPLE_ITEMS_SPECIFIED = 'Multiple {0} are specfied, only one is allowed'
MULTIPLE_ITEMS_MARKED_PRIMARY = 'Multiple {0} are marked primary, only one can be primary'
MULTIPLE_PARENTS_SPECIFIED = 'Multiple parents ({0}) specified, only one is allowed'
MULTIPLE_SEARCH_METHODS_SPECIFIED = 'Multiple search methods ({0}) specified, only one is allowed'
MULTIPLE_SSO_PROFILES_MATCH = 'Multiple SSO profiles match display name {0}:\n'
MULTIPLE_YUBIKEYS_CONNECTED = 'Multiple YubiKeys connected. Specify yubikey_serial_number and one of {0}\n'
MUST_BE_NUMERIC = 'Must be numeric'
NEED_READ_ACCESS = 'Need Read access'
NEED_READ_WRITE_ACCESS = 'Need Read/Write access'
NEED_WRITE_ACCESS = 'Need Write access'
NESTED_LOOP_CMD_NOT_ALLOWED = 'Command can not be nested.'
NEWUSER_REQUIREMENTS = 'newuser option requires: at least 1 recipient and givenname, familyname and password options'
NEW_OWNER_MUST_DIFFER_FROM_OLD_OWNER = 'New owner must differ from old owner'
NO_DATA = 'No data'
NON_BLANK = 'Non-blank'
NON_EMPTY = 'Non-empty'
NOT_A = 'Not a'
NOT_A_PRIMARY_EMAIL_ADDRESS = 'Not a primary email address'
NOT_A_MEMBER = 'Not a member'
NOT_ACTIVE = 'Not Active'
NOT_ALLOWED = 'Not Allowed'
NOT_AN_ENTITY = 'Not a {0}'
NOT_APPROPRIATE = 'Not Appropriate'
NOT_COMPATIBLE = 'Not Compatible'
NOT_COPYABLE = 'Not Copyable'
NOT_COPYABLE_INTO_ITSELF = 'Not copyable into itself'
NOT_COPYABLE_SAME_NAME_CURRENT_FOLDER_MERGE = 'Not copyable with same name into current folder with duplicatefolders merge'
NOT_COPYABLE_SAME_NAME_CURRENT_FOLDER_OVERWRITE = 'Not copyable with same name into current folder with duplicatefiles overwriteall|overwriteolder'
NOT_DELETABLE = 'Not Deletable'
NOT_FOUND = 'Not Found'
NOT_MOVABLE = 'Not Movable'
NOT_MOVABLE_IN_TRASH = 'Not Movable, in Trash'
NOT_MOVABLE_INTO_ITSELF = 'Not movable into itself'
NOT_MOVABLE_SAME_NAME_CURRENT_FOLDER_MERGE = 'Not movable with same name into current folder with duplicatefolders merge'
NOT_MOVABLE_SAME_NAME_CURRENT_FOLDER_OVERWRITE = 'Not movable with same name into current folder with duplicatefiles overwriteall|overwriteolder'
NOT_OWNED_BY = 'Not owned by {0}'
NOT_SELECTED = 'Not Selected'
NOT_WRITABLE = 'Not Writable'
NOW_THE_PRIMARY_DOMAIN = 'Now the primary domain'
NO_ACTION_SPECIFIED = 'No action specified'
NO_AVAILABLE_LICENSES = "There aren't enough available licenses for the specified product-SKU pair(s)"
NO_CHANGES = 'No changes'
NO_CLIENT_ACCESS_ALLOWED = 'No Client Access allowed'
NO_CLIENT_ACCESS_CREATE_UPDATE_ALLOWED = 'No Client Access create/update allowed'
NO_COLUMNS_SELECTED_WITH_CSV_OUTPUT_HEADER_FILTER = 'No columns selected with {0} and {1}'
NO_CREDENTIALS_REPLACEMENT = '{0}: {1} has {2} {3}. We only replace if there are 2.\n'
NO_CSV_DATA_TO_UPLOAD = 'No CSV data to upload'
NO_CSV_FILE_DATA_FOUND = 'No CSV file data found'
NO_CSV_FILE_DATA_SAVED = 'No CSV file data saved'
NO_CSV_FILE_SUBKEYS_SAVED = 'No CSV file subkeys saved'
NO_DATA_TRANSFER_APP_FOR_PARAMETER = 'No data transfer application for key {0}'
NO_ENTITIES_FOUND = 'No {0} found'
NO_ENTITIES_MATCHED = 'No {0} matched'
NO_FILTER_ACTIONS = 'No {0} actions specified'
NO_FILTER_CRITERIA = 'No {0} criteria specified'
NO_LABELS_MATCH = 'No Labels match'
NO_LABELS_TO_PROCESS = 'No Labels to process'
NO_MESSAGES_WITH_LABEL = 'No Messages with Label'
NO_PARENTS_TO_CONVERT_TO_SHORTCUTS = 'No parents to convert to shortcuts'
NO_REPORT_AVAILABLE = 'No {0} report available.'
NO_SCOPES_FOR_API = 'There are no scopes authorized for the {0}'
NO_SERIAL_NUMBERS_SPECIFIED = 'No serial numbers specified'
NO_SSO_PROFILE_MATCHES = 'No SSO profile matches display name {0}'
NO_SSO_PROFILE_ASSIGNED = 'No SSO profile assigned to {0} {1}'
NO_SVCACCT_ACCESS_ALLOWED = 'No Service Account Access allowed'
NO_TRANSFER_LACK_OF_DISK_SPACE = 'Transfer not performed due to lack of target drive space.'
NO_USAGE_PARAMETERS_DATA_AVAILABLE = 'No usage parameters data available.'
NO_USER_COUNTS_DATA_AVAILABLE = 'No User counts data available.'
OAUTH2_GO_TO_LINK_MESSAGE = """
Go to the following link in a browser on this computer or on another computer:

    {url}

If you use a browser on another computer, you will get a browser error that the site can't be reached AFTER you
click the Allow button, paste "Unable to connect" URL from other computer (only URL data up to &scope required):
"""
ON_CURRENT_PRIVATE_KEY = ' on current key'
ON_VAULT_HOLD = 'On Google Vault Hold'
ONLY_ADMINISTRATORS_CAN_PERFORM_SHARED_DRIVE_QUERIES = 'Only administrators can perform Shared Drive queries'
ONLY_ADMINISTRATORS_CAN_SPECIFY_SHARED_DRIVE_ORGUNIT = 'Only administrators can specify Shared Drive Org Unit'
ONLY_ONE_DEVICE_SELECTION_ALLOWED = 'Only one device selection allowed, filter = "{0}"'
ONLY_ONE_JSON_RANGE_ALLOWED = 'Only one range/json allowed'
ONLY_ONE_OWNER_ALLOWED = 'Only one owner allowed'
OR = 'or'
OU_AND_MOVETOOU_CANNOT_BE_IDENTICAL = 'ou {0} can not be be identical to movetoou {1}'
OU_SUBOUS_CANNOT_BE_MOVED_TO_MOVETOOU = 'ou {0} sub OUs can not be be moved to movetoou {1}'
PERMISSION_DENIED = 'The caller does not have permission'
PLEASE_CORRECT_YOUR_SYSTEM_TIME = 'Please correct your system time.'
PLEASE_ENTER_A_OR_M = 'Please enter a or m ...\n'
PLEASE_SELECT_ENTITY_TO_PROCESS = '{0} {1} found, please select the correct one to {2} and specify with {3}'
PLEASE_SPECIFY_BUILDING_EXACT_CASE_NAME_OR_ID = 'Please specify building by exact case name or ID.'
PREVIEW_ONLY = 'Preview Only'
PRIMARY_EMAIL_DID_NOT_MATCH_PATTERN = 'primaryEmail address did not match pattern: {0}'
PROCESS = 'process'
PROCESSES = 'processes'
PROCESSING_ITEM_N = '{0},0,Processing item {1}\n'
PROCESSING_ITEM_N_OF_M = '{0},0,Processing item {1}/{2}\n'
PROFILE_PHOTO_NOT_FOUND = 'Profile photo not found'
PROFILE_PHOTO_IS_DEFAULT = 'Profile photo is default'
REASON_ONLY_VALID_WITH_CONTENTRESTRICTIONS_READONLY_TRUE = 'reason only valid with contentrestrictions readonly true'
REAUTHENTICATION_IS_NEEDED = 'Reauthentication is needed, please run\n\ngam oauth create'
RECOMMEND_RUNNING_GAM_ROTATE_SAKEY = 'Recommend running "gam rotate sakey" to get a new key\n'
REFUSING_TO_DEPROVISION_DEVICES = 'Refusing to deprovision {0} devices because acknowledge_device_touch_requirement not specified.\nDeprovisioning a device means the device will have to be physically wiped and re-enrolled to be managed by your domain again.\nThis requires physical access to the device and is very time consuming to perform for each device.\nPlease add "acknowledge_device_touch_requirement" to the GAM command if you understand this and wish to proceed with the deprovision.\nPlease also be aware that deprovisioning can have an effect on your device license count.\nSee https://support.google.com/chrome/a/answer/3523633 for full details.'
REPLY_TO_CUSTOM_REQUIRES_EMAIL_ADDRESS = 'replyto REPLY_TO_CUSTOM requires customReplyTo <EmailAddress>'
REQUEST_COMPLETED_NO_FILES = 'Request completed but no results/files were returned, try requesting again'
REQUEST_NOT_COMPLETE = 'Request needs to be completed before downloading, current status is: {0}'
RERUN_THE_COMMAND_AND_SPECIFY_A_NEW_SANAME = """
Re-run the command specify a new service account name with: saname <ServiceAccountName>
See: https://github.com/GAM-team/GAM/wiki/Authorization#advanced-use
"""
RESOURCE_CAPACITY_FLOOR_REQUIRED = 'Options "capacity <Number>" (<Number> > 0) and "floor <String>" required'
RESOURCE_FLOOR_REQUIRED = 'Option "floor <String>" required'
RESULTS_TOO_LARGE_FOR_GOOGLE_SPREADSHEET = 'Results are too large for Google Spreadsheets. Uploading as a regular CSV file.'
RETRIES_EXHAUSTED = 'Retries {0} exhausted'
RETRYING_GOOGLE_SHEET_EXPORT_SLEEPING = 'Retrying Google Sheet export {0}/{1}. Sleeping {2} seconds\n'
ROLE_MUST_BE_ORGANIZER = 'Role must be organizer'
ROLE_NOT_IN_SET = 'Role not in set: {0})'
SCHEMA_WOULD_HAVE_NO_FIELDS = '{0} would have no {1}'
SELECTED = 'Selected'
SERVICE_NOT_APPLICABLE = 'Service not applicable/Does not exist'
SERVICE_NOT_APPLICABLE_THIS_ADDRESS = 'Service not applicable for this address: {0}'
SERVICE_NOT_ENABLED = '{0} Service/App not enabled'
SHORTCUT_TARGET_CAPABILITY_IS_FALSE = '{0} capability {1} is False'
SITES_COMMAND_DEPRECATED = 'The Classic Sites API is deprecated, this command will not work:\n{0}'
SKU_HAS_NO_MATCHING_ARCHIVED_USER_SKU = 'SKU {0} has no matching Archived User SKU'
STARTING_THREAD = 'Starting thread'
STATISTICS_COPY_FILE = 'Total: {0}, Copied: {1}, Shortcut created {2}, Shortcut exists {3}, Duplicate: {4}, Copy Failed: {5}, Not copyable: {6}, In skipids: {7}, Permissions Failed: {8}, Protected Ranges Failed: {9}'
STATISTICS_COPY_FOLDER = 'Total: {0}, Copied: {1}, Shortcut created {2}, Shortcut exists {3}, Duplicate: {4}, Merged: {5}, Copy Failed: {6}, Not writable: {7}, Permissions Failed: {8}'
STATISTICS_MOVE_FILE = 'Total: {0}, Moved: {1}, Shortcut created {2}, Shortcut exists {3}, Duplicate: {4}, Move Failed: {5}, Not movable: {6}'
STATISTICS_MOVE_FOLDER = 'Total: {0}, Moved: {1}, Shortcut created {2}, Shortcut exists {3}, Duplicate: {4}, Merged: {5}, Move Failed: {6}, Not writable: {7}'
STATISTICS_USER_NOT_ORGANIZER = 'User not organizer: {0}'
STRING_LENGTH = 'string length'
SUBKEY_FIELD_MISMATCH = 'subkeyfield {0} does not match saved subkeyfield {1}'
SUBSCRIPTION_NOT_FOUND = 'Could not find subscription'
SUFFIX_NOT_ALLOWED_WITH_CUSTOMLANGUAGE = 'Suffix {0} not allowed with customLanguage {1}'
TASKLIST_TITLE_NOT_FOUND = 'Task list title not found'
THREAD = 'thread'
THREADS = 'threads'
TO = 'To'
TO_LC = 'to'
TO_MAXIMUM_OF = 'to maximum of'
TO_SET_UP_GOOGLE_CHAT = """
To set up Google Chat for your current project, please go to:

    {0}

and follow the instructions at:

    https://github.com/GAM-team/GAM/wiki/Chat-Bot#set-up-a-chat-bot

    You'll use projects/{1}/topics/no-topic in Connection settings Cloud Pub/Sub Topic Name
"""
TOTAL_ITEMS_IN_ENTITY = 'Total {0} in {1}'
TRIMMED_MESSAGE_FROM_LENGTH_TO_MAXIMUM = 'Trimmed message of length {0} to maximum length {1}'
UNABLE_TO_GET_PERMISSION_ID = 'Unable to get Permission ID for <{0}>'
UNABLE_TO_CREATE_NOT_FOUND_USER = 'Unable to create not found user, some required field (givenName, familyName, password/notfoundpassword) not present'
UNAVAILABLE = 'Unavailable'
UNKNOWN = 'Unknown'
UNKNOWN_API_OR_VERSION = 'Unknown Google API or version: ({0}), contact {1}'
UNRECOVERABLE_ERROR = 'Unrecoverable error'
UPDATE_ATTENDEE_CHANGES = 'Update attendee changes'
UPDATE_GAM_TO_64BIT = "You're running a 32-bit version of GAM on a 64-bit version of Windows, upgrade to a windows-x86_64 version of GAM"
UPDATE_USER_PASSWORD_CHANGE_NOTIFY_MESSAGE = 'The account password for #givenname# #familyname#, #user# has been changed to: #password#\n'
UPDATE_USER_PASSWORD_CHANGE_NOTIFY_SUBJECT = 'Account #user# password has been changed'
UPLOAD_CSV_FILE_INTERNAL_ERROR = 'Google reported "{0}" but the file was probably uploaded, check that it has {1} rows'
UPLOADING_NEW_PUBLIC_CERTIFICATE_TO_GOOGLE = 'Uploading new public certificate to Google...\n'
URL_ERROR = 'URL error: {0}'
USE_MIMETYPE_TO_SPECIFY_GOOGLE_FORMAT = 'Use "mimetype <MimeType>" to specify Google file format\n'
USED = 'Used'
USER_BELONGS_TO_N_GROUPS_THAT_MAP_TO_ORGUNITS = 'User belongs to {0} groups ({1}) that map to OUs'
USER_CANCELLED = 'User cancelled'
USER_HAS_MULTIPLE_DIRECT_OR_INHERITED_MEMBERSHIPS_IN_GROUP = 'User has multiple direct or inherited memberships in group'
USER_IN_OTHER_DOMAIN = '{0}: {1} in other domain.'
USER_IS_NOT_ORGANIZER = 'User is not organizer, use anyorganizer option to override'
USER_NOT_IN_MATCHUSERS = 'User not in matchusers'
USER_SUBS_NOT_ALLOWED_TAG_REPLACEMENT = 'user substitutions not allowed in replace <Tag> <String>'
USE_DOIT_ARGUMENT_TO_PERFORM_ACTION = 'Use the "doit" argument to perform action'
USING_N_PROCESSES = '{0},0/{1},Using {2} {3}...\n'
VALUES_ARE_NOT_CONSISTENT = 'Values are not consistent'
VERSION_UPDATE_AVAILABLE = 'Version update available'
WAITING_FOR_DATA_TRANSFER_TO_COMPLETE_SLEEPING = 'Waiting for Data Transfer to complete. Sleeping {0} seconds\n'
WAITING_FOR_ITEM_CREATION_TO_COMPLETE_SLEEPING = 'Waiting for {0} creation to complete. Sleeping {1} seconds\n'
WHAT_IS_YOUR_PROJECT_ID = '\nWhat is your project ID? '
WILL_RERUN_WITH_NO_BROWSER_TRUE = 'Will re-run command with no_browser true\n'
WITH = 'with'
WOULD_MAKE_MEMBERSHIP_CYCLE = 'Would make membership cycle'
WRITER_ACCESS_REQUIRED_TO_BOTH_CALENDARS = 'Writer access required to both calendars'
WROTE_PRIVATE_KEY_DATA = 'Wrote private key data to {0}\n'
WROTE_PUBLIC_CERTIFICATE = 'Wrote public certificate to {0}\n'
YOU_CAN_ADD_DOMAIN_TO_ACCOUNT = 'You can now add: {0} or its subdomains as secondary or domain aliases of the Google Workspace Account: {1}'
YUBIKEY_GENERATING_NONEXPORTABLE_PRIVATE_KEY = 'YubiKey is generating a non-exportable private key...\n'
YUBIKEY_PIN_SET_TO = 'YubiKey PIN set to: {0}\n'



#!/usr/bin/env python
#
# Copyright (C) 2009 Google Inc.
#
# Licensed under the Apache License 2.0;



"""AtomPubClient provides CRUD ops. in line with the Atom Publishing Protocol.

"""

# __author__ = 'j.s@google.com (Jeff Scudder)'

import atom.http_core


class Error(Exception):
    pass


class MissingHost(Error):
    pass


class AtomPubClient(object):
    host = None
    auth_token = None
    ssl = False  # Whether to force all requests over https
    xoauth_requestor_id = None

    def __init__(self, http_client=None, host=None, auth_token=None, source=None,
                 xoauth_requestor_id=None, **kwargs):
        """Creates a new AtomPubClient instance.

        Args:
          source: The name of your application.
          http_client: An object capable of performing HTTP requests through a
                       request method. This object is used to perform the request
                       when the AtomPubClient's request method is called. Used to
                       allow HTTP requests to be directed to a mock server, or use
                       an alternate library instead of the default of httplib to
                       make HTTP requests.
          host: str The default host name to use if a host is not specified in the
                requested URI.
          auth_token: An object which sets the HTTP Authorization header when its
                      modify_request method is called.
        """
        self.http_client = http_client or atom.http_core.ProxiedHttpClient()
        if host is not None:
            self.host = host
        if auth_token is not None:
            self.auth_token = auth_token
        self.xoauth_requestor_id = xoauth_requestor_id
        self.source = source

    def request(self, method=None, uri=None, auth_token=None,
                http_request=None, **kwargs):
        """Performs an HTTP request to the server indicated.

        Uses the http_client instance to make the request.

        Args:
          method: The HTTP method as a string, usually one of 'GET', 'POST',
                  'PUT', or 'DELETE'
          uri: The URI desired as a string or atom.http_core.Uri.
          http_request:
          auth_token: An authorization token object whose modify_request method
                      sets the HTTP Authorization header.

        Returns:
          The results of calling self.http_client.request. With the default
          http_client, this is an HTTP response object.
        """
        # Modify the request based on the AtomPubClient settings and parameters
        # passed in to the request.
        http_request = self.modify_request(http_request)
        if isinstance(uri, str):
            uri = atom.http_core.Uri.parse_uri(uri)
        if uri is not None:
            uri.modify_request(http_request)
        if isinstance(method, str):
            http_request.method = method
        # Any unrecognized arguments are assumed to be capable of modifying the
        # HTTP request.
        for name, value in kwargs.items():
            if value is not None:
                if hasattr(value, 'modify_request'):
                    value.modify_request(http_request)
                else:
                    http_request.uri.query[name] = str(value)
        # Default to an http request if the protocol scheme is not set.
        if http_request.uri.scheme is None:
            http_request.uri.scheme = 'http'
        # Override scheme. Force requests over https.
        if self.ssl:
            http_request.uri.scheme = 'https'
        if http_request.uri.path is None:
            http_request.uri.path = '/'
        # Add the Authorization header at the very end. The Authorization header
        # value may need to be calculated using information in the request.
        if auth_token:
            auth_token.modify_request(http_request)
        elif self.auth_token:
            self.auth_token.modify_request(http_request)
        # Check to make sure there is a host in the http_request.
        if http_request.uri.host is None:
            raise MissingHost('No host provided in request %s %s' % (
                http_request.method, str(http_request.uri)))
        # Perform the fully specified request using the http_client instance.
        # Sends the request to the server and returns the server's response.
        return self.http_client.request(http_request)

    Request = request

    def get(self, uri=None, auth_token=None, http_request=None, **kwargs):
        """Performs a request using the GET method, returns an HTTP response."""
        return self.request(method='GET', uri=uri, auth_token=auth_token,
                            http_request=http_request, **kwargs)

    Get = get

    def post(self, uri=None, data=None, auth_token=None, http_request=None,
             **kwargs):
        """Sends data using the POST method, returns an HTTP response."""
        return self.request(method='POST', uri=uri, auth_token=auth_token,
                            http_request=http_request, data=data, **kwargs)

    Post = post

    def put(self, uri=None, data=None, auth_token=None, http_request=None,
            **kwargs):
        """Sends data using the PUT method, returns an HTTP response."""
        return self.request(method='PUT', uri=uri, auth_token=auth_token,
                            http_request=http_request, data=data, **kwargs)

    Put = put

    def delete(self, uri=None, auth_token=None, http_request=None, **kwargs):
        """Performs a request using the DELETE method, returns an HTTP response."""
        return self.request(method='DELETE', uri=uri, auth_token=auth_token,
                            http_request=http_request, **kwargs)

    Delete = delete

    def modify_request(self, http_request):
        """Changes the HTTP request before sending it to the server.

        Sets the User-Agent HTTP header and fills in the HTTP host portion
        of the URL if one was not included in the request (for this it uses
        the self.host member if one is set). This method is called in
        self.request.

        Args:
          http_request: An atom.http_core.HttpRequest() (optional) If one is
                        not provided, a new HttpRequest is instantiated.

        Returns:
          An atom.http_core.HttpRequest() with the User-Agent header set and
          if this client has a value in its host member, the host in the request
          URL is set.
        """
        if http_request is None:
            http_request = atom.http_core.HttpRequest()

        if self.host is not None and http_request.uri.host is None:
            http_request.uri.host = self.host

        if self.xoauth_requestor_id is not None:
            http_request.uri.query['xoauth_requestor_id'] = self.xoauth_requestor_id

        # Set the user agent header for logging purposes.
        if self.source:
            http_request.headers['User-Agent'] = '%s gdata-py/2.0.18' % self.source
        else:
            http_request.headers['User-Agent'] = 'gdata-py/2.0.17'

        return http_request

    ModifyRequest = modify_request


class CustomHeaders(object):
    """Add custom headers to an http_request.

    Usage:
      >>> custom_headers = atom.client.CustomHeaders(header1='value1',
              header2='value2')
      >>> client.get(uri, custom_headers=custom_headers)
    """

    def __init__(self, **kwargs):
        """Creates a CustomHeaders instance.

        Initialize the headers dictionary with the arguments list.
        """
        self.headers = kwargs

    def modify_request(self, http_request):
        """Changes the HTTP request before sending it to the server.

        Adds the custom headers to the HTTP request.

        Args:
          http_request: An atom.http_core.HttpRequest().

        Returns:
          An atom.http_core.HttpRequest() with the added custom headers.
        """

        for name, value in self.headers.items():
            if value is not None:
                http_request.headers[name] = value
        return http_request



# Copyright 2014 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Caching utility for the discovery document."""

from __future__ import absolute_import

import logging
import os

LOGGER = logging.getLogger(__name__)

DISCOVERY_DOC_MAX_AGE = 60 * 60 * 24  # 1 day
DISCOVERY_DOC_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "documents"
)


def autodetect():
    """Detects an appropriate cache module and returns it.

    Returns:
      googleapiclient.discovery_cache.base.Cache, a cache object which
      is auto detected, or None if no cache object is available.
    """
    if "GAE_ENV" in os.environ:
        try:
            from . import appengine_memcache

            return appengine_memcache.cache
        except Exception:
            pass
    try:
        from . import file_cache

        return file_cache.cache
    except Exception:
        LOGGER.info(
            "file_cache is only supported with oauth2client<4.0.0", exc_info=False
        )
        return None


def get_static_doc(serviceName, version):
    """Retrieves the discovery document from the directory defined in
    DISCOVERY_DOC_DIR corresponding to the serviceName and version provided.

    Args:
        serviceName: string, name of the service.
        version: string, the version of the service.

    Returns:
        A string containing the contents of the JSON discovery document,
        otherwise None if the JSON discovery document was not found.
    """

    content = None
    doc_name = "{}.{}.json".format(serviceName, version)

    try:
        with open(os.path.join(DISCOVERY_DOC_DIR, doc_name), "r") as f:
            content = f.read()
    except FileNotFoundError:
        # File does not exist. Nothing to do here.
        pass

    return content



# -*- coding: utf-8 -*-
"""ISO 8601 date time string parsing

"""

from datetime import (datetime, timedelta, tzinfo)
import time as _time
import re

ISO8601_REGEX = re.compile(r'^(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})(?P<separator>[ T])(?P<hour>[0-9]{2}):(?P<minute>[0-9]{2}):(?P<second>[0-9]{2})([.,](?P<second_fraction>[0-9]+)){0,1}(?P<timezone>Z|(?P<tz_sign>[-+])(?P<tz_hour>[0-9]{2}):(?P<tz_minute>[0-9]{2}))$')
ISO8601_TZ_REGEX = re.compile(r'^(?P<timezone>Z|(?P<tz_sign>[-+])(?P<tz_hour>[0-9]{2}):(?P<tz_minute>[0-9]{2}))$')

class ParseError(Exception):
  """Raised when there is a problem parsing a date string"""

# Yoinked from python docs
ZERO = timedelta(0)
class Utc(tzinfo):
  """UTC Timezone

  """
  def utcoffset(self, dt):
    return ZERO

  def tzname(self, dt):
    return "UTC"

  def dst(self, dt):
    return ZERO

  def __repr__(self):
    return "<iso8601.Utc>"

UTC = Utc()

class FixedOffset(tzinfo):
  """Fixed offset in hours and minutes from UTC

  """
  def __init__(self, offset_hours, offset_minutes, name):
    self.__offset_hours = offset_hours  # Keep for later __getinitargs__
    self.__offset_minutes = offset_minutes  # Keep for later __getinitargs__
    self.__offset = timedelta(hours=offset_hours, minutes=offset_minutes)
    self.__name = name

  def __eq__(self, other):
    if isinstance(other, FixedOffset):
      return (other.__offset == self.__offset) and (other.__name == self.__name)
    if isinstance(other, tzinfo):
      return other == self
    return False

  def __getinitargs__(self):
    return (self.__offset_hours, self.__offset_minutes, self.__name)

  def utcoffset(self, dt):
    return self.__offset

  def tzname(self, dt):
    return self.__name

  def dst(self, dt):
    return ZERO

  def __repr__(self):
    return "<FixedOffset %r %r>" % (self.__name, self.__offset)

# A class capturing the platform's idea of local time.

STDOFFSET = timedelta(seconds = -_time.timezone)
if _time.daylight:
  DSTOFFSET = timedelta(seconds = -_time.altzone)
else:
  DSTOFFSET = STDOFFSET

DSTDIFF = DSTOFFSET - STDOFFSET

class LocalTimezone(tzinfo):
  """Local time zone

  """

  def utcoffset(self, dt):
    if self._isdst(dt):
      return DSTOFFSET
    else:
      return STDOFFSET

  def dst(self, dt):
    if self._isdst(dt):
      return DSTDIFF
    else:
      return ZERO

  def tzname(self, dt):
    return _time.tzname[self._isdst(dt)]

  def _isdst(self, dt):
    tt = (dt.year, dt.month, dt.day,
          dt.hour, dt.minute, dt.second,
          dt.weekday(), 0, 0)
    stamp = _time.mktime(tt)
    tt = _time.localtime(stamp)
    return tt.tm_isdst > 0

Local = LocalTimezone()

def parse_timezone(matches):
  """Parses ISO 8601 time zone specs into tzinfo offsets

  """

  if matches["timezone"] == "Z":
    return UTC
  sign = matches["tz_sign"]
  hours = int(matches['tz_hour'])
  minutes = int(matches['tz_minute'])
  description = "%s%02d:%02d" % (sign, hours, minutes)
  if sign == "-":
    hours = -hours
    minutes = -minutes
  return FixedOffset(hours, minutes, description)

def parse_timezone_str(tzstring):
  m = ISO8601_TZ_REGEX.match(tzstring)
  if not m:
    raise ParseError("Unable to parse timezone string %r" % tzstring)
  groups = m.groupdict()
  return parse_timezone(groups)

def parse_date(datestring):
  """Parses ISO 8601 dates into datetime objects

  The timezone is parsed from the date string. However it is quite common to
  have dates without a timezone (not strictly correct). In this case the
  default timezone specified in default_timezone is used. This is UTC by
  default.

  :param datestring: The date to parse as a string
  :returns: A datetime.datetime instance
  :raises: ParseError when there is a problem parsing the date or
           constructing the datetime instance.

  """
  m = ISO8601_REGEX.match(datestring)
  if not m:
    raise ParseError("Unable to parse date string %r" % datestring)
  groups = m.groupdict()
  tz = parse_timezone(groups)
  try:
    return (datetime(year=int(groups['year']),
                     month=int(groups['month']),
                     day=int(groups['day']),
                     hour=int(groups['hour']),
                     minute=int(groups['minute']),
                     second=int(groups['second']),
                     tzinfo=tz),
            tz)
  except Exception as e:
    raise ParseError(e)
